# coding: utf-8

from struct import pack_into
from time import sleep
from typing import Union
from .actuator import Actuator
from .._global import OptionalModule
from ..tool import ft232h_server as ft232h, Usb_server

try:
  from adafruit_motorkit import MotorKit
except (ImportError, ModuleNotFoundError):
  MotorKit = OptionalModule('adafruit_motorkit',
                            'Adafrfuit Motorkit module (adafruit_motorkit) is '
                            'required to use this actuator')

try:
  import board
except (ImportError, ModuleNotFoundError):
  board = OptionalModule('board', 'Blinka is necessary to use the I2C bus')

try:
  from smbus2 import SMBus
except (ImportError, ModuleNotFoundError):
  SMBus = OptionalModule('smbus2')

motor_hat_ctrl = {1: 0x26,
                  2: 0x3A,
                  3: 0x0E,
                  4: 0x22}

motor_hat_pos = {1: 0x2A,
                 2: 0x32,
                 3: 0x12,
                 4: 0x1A}

motor_hat_neg = {1: 0x2E,
                 2: 0x36,
                 3: 0x16,
                 4: 0x1E}

motor_hat_0xFF = [0x00, 0x10, 0x00, 0x00]
motor_hat_backends = ['Pi4', 'blinka', 'ft232h']
motor_hat_max_volt = 12


class DC_motor_hat:
  """Class for driving Adafruit's DC motor HAT.

  This class serves as a basis for building Actuators in Crappy, but is not one
  itself. It is used by the :class:`Motorkit_pump` Actuator.

  Note:
    This device can also drive stepper motors, but this feature isn't included
    here.

  It is intended for Raspberry Pis but can also be used from any other device
  interfacing over I2C assuming a proper wiring.
  """

  def __init__(self,
               motor_nrs: list,
               device_address: int = 0x60,
               i2c_port: int = 1,
               bus: ft232h = None) -> None:
    """Resets the HAT and initializes it.

    Args:
      motor_nrs (:obj:`list`): The list of the motors to drive. Its elements
        should be integers between 1 and 4, corresponding to the number of the
        motors.
      device_address (:obj:`int`, optional): The I2C address of the HAT.
        The default address is `0x60`, but it is possible to change this
        setting by cutting traces on the board.
      i2c_port (:obj:`int`, optional): The I2C port over which the ADS1115
        should communicate. On most Raspberry Pi models the default I2C port is
        `1`.
      bus (:obj:`ft232h_server`, optional): If given, the I2C commands are sent
        by the corresponding :class:`ft232h_server` instance, in the situation
        when the hat is controlled from a PC through an FT232H rather than by a
        board.
    """

    if not all(i in range(1, 5) for i in motor_nrs):
      raise ValueError("The DC motor hat can only drive up to 4 DC motors at "
                       "a time !")

    if not isinstance(device_address, int):
      raise TypeError("device_address should be an integer between 0 and 127.")
    self._address = device_address

    if not isinstance(i2c_port, int):
      raise TypeError("i2c_port should be an integer !")

    if not bus:
      self._bus = SMBus(i2c_port)
    else:
      self._bus = bus

    self._buf = bytearray(4)

    # Reset
    self._write_i2c(0x00, [0x00])
    sleep(0.01)

    # Sleep & restart, also setting the frequency to 1526Hz
    self._write_i2c(0x00, [0x10])
    prescale = int(25E6 / 4096.0 / 1600.0 + 0.5)
    self._write_i2c(0xFE, [prescale])
    sleep(0.01)
    self._write_i2c(0x00, [0x00])
    sleep(0.01)
    self._write_i2c(0x00, [0xA0])

    # Initializing the motors
    for i in motor_nrs:
      self._write_i2c(motor_hat_ctrl[i], motor_hat_0xFF)

  def set_motor(self, nr: int, cmd: float) -> None:
    """Sets the PWMs associated with a given motor.

    Args:
      nr (:obj:`int`): The number of the motor to drive.
      cmd (:obj:`float`): The command, that should be between -1 and 1.
    """

    # Validity checks
    if nr not in range(1, 5):
      raise ValueError("It is only possible to drive up to 4 motors.")
    if not -1.0 <= cmd <= 1.0:
      raise ValueError("The command for the motor should be between -1 and 1.")

    # Special settings for the extreme values
    if cmd == 0:
      self._write_i2c(motor_hat_pos[nr], motor_hat_0xFF)
      self._write_i2c(motor_hat_neg[nr], motor_hat_0xFF)
    elif cmd == 1.0:
      self._write_i2c(motor_hat_pos[nr], motor_hat_0xFF)
      self._write_i2c(motor_hat_neg[nr], [0x00 for _ in range(4)])
    elif cmd == -1.0:
      self._write_i2c(motor_hat_pos[nr], [0x00 for _ in range(4)])
      self._write_i2c(motor_hat_neg[nr], motor_hat_0xFF)

    # The positive line is driven for positive commands, and reversely
    elif cmd > 0:
      duty_cycle = (int(0xFFFF * cmd) + 1) >> 4
      pack_into("<HH", self._buf, 0, *(0, duty_cycle))
      self._write_i2c(motor_hat_pos[nr], self._buf)
      self._write_i2c(motor_hat_neg[nr], [0x00 for _ in range(4)])
    else:
      duty_cycle = (int(0xFFFF * abs(cmd)) + 1) >> 4
      pack_into("<HH", self._buf, 0, *(0, duty_cycle))
      self._write_i2c(motor_hat_pos[nr], [0x00 for _ in range(4)])
      self._write_i2c(motor_hat_neg[nr], self._buf)

  def close(self) -> None:
    """Closes the I2C bus."""

    self._bus.close()

  def _write_i2c(self, register: int, buf: Union[list, bytearray]):
    """Thin wrapper to reduce verbosity."""

    self._bus.write_i2c_block_data(self._address, register, list(buf))


class Motorkit_pump(Usb_server, Actuator):
  """Class for controlling two DC air pumps and a valve.

  It uses Adafruit's DC motor HAT. The motor 1 controls the inflation pump,
  the motor 2 controls a valve, and the motor 3 controls a deflation pump.
  """

  def __init__(self,
               backend: str,
               device_address: int = 0x60,
               i2c_port: int = 1,
               ft232h_ser_num: str = None) -> None:
    """Checks the validity of the arguments.

    Args:
      backend (:obj:`str`): Should be one of :
        ::

          'Pi4', 'blinka', 'ft232h'

        The `'Pi4'` backend is optimized but only works on boards supporting
        the :mod:`smbus2` module, like the Raspberry Pis. The `'blinka'`
        backend may be less performant and requires installing Adafruit's
        modules, but these modules are compatible with and maintained on a wide
        variety of boards. The `'ft232h'` backend allows controlling the hat
        from a PC using Adafruit's FT232H USB to I2C adapter. See
        :ref:`Crappy for embedded hardware` for details.
      device_address (:obj:`int`, optional): The I2C address of the HAT.
        The default address is `0x60`, but it is possible to change this
        setting by cutting traces on the board.
      i2c_port (:obj:`int`, optional): The I2C port over which the HAT should
        communicate. On most Raspberry Pi models the default I2C port is
        `1`.
      ft232h_ser_num (:obj:`str`, optional): If backend is `'ft232h'`, the
        serial number of the ft232h to use for communication.
    """

    if not isinstance(backend, str) or backend not in motor_hat_backends:
      raise ValueError("backend should be in {}".format(motor_hat_backends))
    self._backend = backend

    Usb_server.__init__(self,
                        serial_nr=ft232h_ser_num if ft232h_ser_num else '',
                        backend=backend)
    Actuator.__init__(self)
    current_file, block_number, command_file, answer_file, block_lock, \
        current_lock = super().start_server()

    if self._backend == 'ft232h':
      self._bus = ft232h(mode='I2C',
                         block_number=block_number,
                         current_file=current_file,
                         command_file=command_file,
                         answer_file=answer_file,
                         block_lock=block_lock,
                         current_lock=current_lock,
                         serial_nr='')
    else:
      self._bus = None

    if not isinstance(device_address, int):
      raise TypeError("device_address should be an integer between 0 and 127.")
    self._address = device_address

    if not isinstance(i2c_port, int):
      raise TypeError("i2c_port should be an integer !")
    self._port = i2c_port

  def open(self) -> None:
    """Initializes the generic HAT object."""

    if self._backend == 'blinka':
      self._hat = MotorKit(i2c=board.I2C())

      def set_motor(nr: int, cmd: float) -> None:
        setattr(getattr(self._hat, 'motor{}'.format(nr)), 'throttle', cmd)

      self._hat.set_motor = set_motor

    else:
      self._hat = DC_motor_hat([1, 2, 3], self._address, self._port, self._bus)

  def set_speed(self, volt: float) -> None:
    """Inflates or deflates the setup according to the command.

    Args:
      volt (:obj:`float`): The voltage to supply to the pumps. If positive,
        will inflate, if negative will deflate, and if 0 won't do anything. The
        voltage is clamped between -12 and 12 Volts, as it is the limit of the
        HAT.
    """

    volt_clamped = min(abs(volt) / motor_hat_max_volt, 1.0)

    # Stops all the motors
    if volt == 0:
      self._hat.set_motor(1, 0.0)
      self._hat.set_motor(2, 0.0)
      self._hat.set_motor(3, 0.0)

    # Drives the inflating pump
    elif volt > 0:
      self._hat.set_motor(1, volt_clamped)
      self._hat.set_motor(2, 0.0)
      self._hat.set_motor(3, 0.0)

    # Drives the deflating pump and opens the valve
    elif volt < 0:
      self._hat.set_motor(1, 0.0)
      self._hat.set_motor(2, 1.0)
      self._hat.set_motor(3, volt_clamped)

  def stop(self) -> None:
    """Simply sets the voltage to 0."""

    self.set_speed(0.0)

  def close(self) -> None:
    """Stops the pumps and closes the HAT object."""

    self.stop()
    if self._backend != 'blinka':
      self._hat.close()



# coding: utf-8

from multiprocessing import Lock
from typing import Union

from .actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class Servostar(Actuator):
  """To drive and configure a servostar variator through a serial
  connection."""

  def __init__(self,
               device: str,
               baudrate: int = 38400,
               mode: str = "serial") -> None:
    """Sets the instance attributes.

    Args:
      device (:obj:`str`): Path to connect to the serial port.
      baudrate (:obj:`int`, optional): Set the corresponding baud rate.
      mode (:obj:`str`, optional): Can be `'analog'` or `'serial'`.
    """

    Actuator.__init__(self)
    self.devname = device
    self.mode = mode
    self.baud = baudrate
    self.lock = Lock()
    self.last = None

  def open(self) -> None:
    self.lock.acquire()
    self.ser = serial.Serial(self.devname, baudrate=self.baud, timeout=2)
    self.ser.flushInput()
    self.ser.write('ANCNFG 0\r\n')
    self.lock.release()
    if self.mode == "analog":
      self.set_mode_analog()
    elif self.mode == "serial":
      self.set_mode_serial()
    else:
      raise AttributeError("No such mode: " + str(self.mode))
    self.lock.acquire()
    self.ser.write('EN\r\n')
    self.ser.write('MH\r\n')
    self.lock.release()

  def set_position(self,
                   pos: float,
                   speed: float = 20000,
                   acc: float = 200,
                   dec: float = 200) -> None:
    """Go to the position specified at the given speed and acceleration."""

    if self.last is pos:
      return
    if isinstance(pos, bool):
      # To use set_position(True) as set_mode_serial()
      # and set_position(False) as set_mode_analog()
      # (to command all of this from as single generator)
      if pos:
        self.set_mode_serial()
      else:
        self.set_mode_analog()
    elif self.mode != "serial":
      self.set_mode_serial()
    self.lock.acquire()
    self.ser.flushInput()
    self.ser.write(" ".join(["ORDER 0", str(int(pos)), str(speed),
                   "8192", str(acc), str(dec), "0 0 0 0\r\n"]))
    self.ser.write("MOVE 0\r\n")  # activates the order
    self.lock.release()
    self.last = pos

  def get_position(self) -> Union[float, None]:
    """Reads current position.

    Returns:
      Current position of the motor.
    """

    self.lock.acquire()
    self.ser.flushInput()
    self.ser.write("PFB\r\n")
    r = ''
    while r != "PFB\r\n":
      if len(r) == 5:
        r = r[1:]
      r += self.ser.read()
      if not r:
        print("Servostar timeout error! make sure the servostar is on!")
        self.lock.release()
        return
    r = ''
    while "\n" not in r:
      r += self.ser.read()
    self.lock.release()
    return int(r)

  def set_mode_serial(self) -> None:
    """Sets the serial input as setpoint."""

    self.lock.acquire()
    self.ser.flushInput()
    self.ser.write('OPMODE 8\r\n')
    self.lock.release()
    self.mode = "serial"

  def set_mode_analog(self) -> None:
    """Sets the analog input as setpoint."""

    self.last = None
    self.lock.acquire()
    self.ser.flushInput()
    self.ser.write('OPMODE 1\r\n')
    self.lock.release()
    self.mode = "analog"

  def clear_errors(self) -> None:
    """Clears error in motor registers."""

    self.ser.flushInput()
    self.ser.write("CLRFAULT\r\n")

  def stop(self) -> None:
    """Stops the motor."""

    self.ser.write("DIS\r\n")
    self.ser.flushInput()

  def close(self) -> None:
    self.ser.close()



# coding: utf-8

from time import time
from numpy import sin, pi
from typing import Union, Dict

from .path import Path, condition_type


class Sine(Path):
  """This path generates a sine wave varying with time until the stop condition
  is met."""

  def __init__(self,
               _last_time: float,
               _last_cmd: float,
               condition: Union[str, condition_type],
               freq: float,
               amplitude: float,
               offset: float = 0,
               phase: float = 0) -> None:
    """Sets the args and initializes the parent class.

    Args:
      _last_time: The last timestamp when a command was generated. For internal
        use only, do not overwrite.
      _last_cmd: The last sent command. For internal use only, do not
        overwrite.
      condition: The condition for switching to the next path. Refer to
        :ref:`generator path` for more info.
      freq: The frequency of the sine in `Hz`.
      amplitude: The amplitude of the sine wave (peak to peak).
      offset: The offset of the sine (average value).
      phase: The phase of the sine (in radians).
    """

    Path.__init__(self, _last_time, _last_cmd)

    # Setting the attributes
    self._condition = self.parse_condition(condition)
    self._amplitude = amplitude / 2
    self._offset = offset
    self._phase = phase
    self._k = 2 * pi * freq

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Returns the value to send or raises :exc:`StopIteration` if the stop
    condition is met."""

    # Checking if the stop condition is met
    if self._condition(data):
      raise StopIteration

    # Returning the current signal value
    return sin((time() - self.t0) * self._k - self._phase) * \
        self._amplitude + self._offset



# coding: utf-8

from time import sleep
from typing import List, Optional, Dict, Any, Union
from pathlib import Path

from .block import Block


class Recorder(Block):
  """Saves data from an upstream block to a text file, with values separated by
  a coma and lines by a newline character.

  The first row of the file contains the names of the saved labels.
  This block can only save data coming from one upstream block. To save data
  from multiple blocks, use several instances of Recorder (recommended) or a
  :ref:`Multiplex` block.
  """

  def __init__(self,
               filename: Union[str, Path],
               delay: float = 2,
               labels: Optional[List[str]] = None,
               freq: float = 200,
               verbose: bool = True) -> None:
    """Sets the args and initializes the parent class.

    Args:
      filename: Path to the output file, either relative or absolute. If the
        parent folders of the file do not exist, they will be created. If the
        file already exists, the actual file where data will be written will be
        renamed with a trailing index to avoid overriding it.
      delay: Delay between each write in seconds.
      labels: If provided, only the data carried by these labels will be saved.
        Otherwise, all the received data is saved.
      freq: The block will try to loop at this frequency.
      verbose: If :obj:`True`, prints the looping frequency of the block.
    """

    super().__init__()
    self.niceness = -5
    self.freq = freq
    self.verbose = verbose

    self._delay = delay
    self._path = Path(filename)
    self._labels = labels

  def prepare(self) -> None:
    """Checking that the block has the right number of inputs, creates the
    folder containing the file if it doesn't already exist, and changes the
    name of the file if it already exists."""

    # Making sure there's the right number of incoming links
    if not self.inputs:
      raise ValueError('The Recorder block does not have inputs !')
    elif len(self.inputs) > 1:
      raise ValueError('Cannot link more than one block to a Recorder block !')
    self._link = self.inputs[0]

    parent_folder = self._path.parent

    # Creating the folder for storing the data if it does not already exist
    if not Path.is_dir(parent_folder):
      Path.mkdir(parent_folder, exist_ok=True, parents=True)

    # Changing the name of the file if it already exists
    if Path.exists(self._path):
      print(f'[Recorder] Warning ! The file {self._path} already exists !')
      stem, suffix = self._path.stem, self._path.suffix
      i = 1
      # Adding an integer at the end of the name to identify the file
      while Path.exists(parent_folder / f'{stem}_{i:05d}{suffix}'):
        i += 1
      self._path = parent_folder / f'{stem}_{i:05d}{suffix}'
      print(f'[Recorder] Using {self._path} instead !')

  def begin(self) -> None:
    """Receives the first chunk of data, writes the labels names in the first
    row of the file and starts saving the actual data."""

    data = self._link.recv_delay(self._delay)

    # If no labels are given, save everything that's received
    if self._labels is None:
      self._labels = list(data.keys())

    # The first row of the file contains the names of the labels
    with open(self._path, 'w') as file:
      file.write(f"{','.join(self._labels)}\n")

    # The following rows contain data
    self._save(data)

  def loop(self) -> None:
    """Simply receives data from the upstream block and saves it."""

    self._save(self._link.recv_delay(self._delay))

  def finish(self) -> None:
    """Gathers any data left in the links, and saves it."""

    sleep(0.5)
    data = self._link.recv_chunk_no_stop()
    if data is not None:
      self._save(data)

  def _save(self, data: Dict[str, List[Any]]):
    """Saves only the data carried by the specified labels to the file."""

    # Keeping only the data that needs to be saved
    data = {key: val for key, val in data.items() if key in self._labels}

    with open(self._path, 'a') as file:
      # Sorting the lists of values in the same order as the labels
      sorted_data = [data[label] for label in self._labels]
      # Actually writing the values
      for values in zip(*sorted_data):
        file.write(f"{','.join(map(str, values))}\n")


class Saver(Recorder):
  def __init__(self, *args, **kwargs) -> None:
    print('#### WARNING ####\n'
          'The block "Saver" has been renamed to "Recorder".\n'
          'Please replace the name in your program, '
          'it will be removed in future versions\n'
          '#################')
    super().__init__(*args, **kwargs)



# coding: utf-8

from crappy.tool import ft232h

"""
Utility for writing a serial number to an FT32H.

Simply run this program, it will ask you for the serial number to write and 
will then write it.
"""

serial_number = None

while not isinstance(serial_number, int):
  try:
    serial_number = int(input("Please enter the desired serial number : "))
  except ValueError:
    print("Wrong input, serial_number should be an integer !")
    print("")

FT232H = ft232h('Write_serial_nr', str(serial_number))

print("Serial number successfully written")



# coding: utf-8

"""
Demonstration of a DIC controlled test.

This program is intended as a demonstration and is fully virtual.

No hardware required.
Requires the cv2 module to be installed.
"""

import crappy
import numpy as np
import cv2


class Apply_strain_img:
  """This class reshapes an image depending on input strain values."""

  def __init__(self,
               image: np.ndarray) -> None:
    """Sets the base image and initializes the necessary objects to use."""

    self._img = image

    # Building the lookup arrays for the cv2.remap method
    height, width, *_ = image.shape
    orig_x, orig_y = np.meshgrid(range(width), range(height))
    # These arrays correspond to the original state of the image
    self._orig_x = orig_x.astype(np.float32)
    self._orig_y = orig_y.astype(np.float32)

    # These arrays are meant to be added to the original image ones
    # If added as is, they correspond to a 100% strain state in both directions
    self._x_strain = self._orig_x * width / (width - 1) - width / 2
    self._y_strain = self._orig_y * height / (height - 1) - height / 2

  def __call__(self, exx: float, eyy: float) -> np.ndarray:
    """Returns the reshaped image, based on the given strain values."""

    exx /= 100
    eyy /= 100

    # The final lookup table is the sum of the original state ones plus the
    # 100% strain one weighted by a ratio
    transform_x = self._orig_x - (exx / (1 + exx)) * self._x_strain
    transform_y = self._orig_y - (eyy / (1 + eyy)) * self._y_strain

    return cv2.remap(self._img, transform_x, transform_y, 1)


if __name__ == "__main__":
  img = crappy.resources.speckle

  speed = 5 / 60  # mm/sec

  # Load until the strain is reached, then unload until force is 0
  generator = crappy.blocks.Generator(path=sum([[
    {'type': 'constant', 'value': speed, 'condition': 'Exx(%)>{}'.format(5*i)},
    {'type': 'constant', 'value': -speed, 'condition': 'F(N)<0'}]
    for i in range(1, 5)], []), spam=False)

  # Our fake machine
  machine = crappy.blocks.Fake_machine(max_strain=17, k=5000, l0=20,
                                       plastic_law=lambda exx: 0,
                                       sigma={'F(N)': 0.5})

  crappy.link(generator, machine)
  crappy.link(machine, generator)

  # The block performing the DIC
  dis = crappy.blocks.DISCorrel('', show_image=True,
                                labels=['t(s)', 'x', 'y', 'measured_Exx(%)',
                                        'measured_Eyy(%)'],
                                image_generator=Apply_strain_img(img),
                                verbose=True, display_images=True)
  # This modifier will generate an image with the values of strain
  # coming from the Fake_machine block
  crappy.link(machine, dis)

  graph_def2 = crappy.blocks.Grapher(('t(s)', 'measured_Exx(%)'),
                                     ('t(s)', 'measured_Eyy(%)'))
  crappy.link(dis, graph_def2)

  crappy.start()



# coding: utf-8

from .._global import DefinitionError


class MetaModifier(type):
  """Metaclass keeping track of all the Modifiers, including the custom
  user-defined ones."""

  classes = {}

  needed_methods = ["evaluate"]

  def __new__(mcs, name: str, bases: tuple, dct: dict) -> type:
    return super().__new__(mcs, name, bases, dct)

  def __init__(cls, name: str, bases: tuple, dct: dict) -> None:
    super().__init__(name, bases, dct)

    # Checking that a Modifier with the same name doesn't already exist
    if name in cls.classes or name == "Modifier":
      raise DefinitionError(f"The {name} class is already defined !")

    # Gathering all the defined methods
    defined_methods = list(dct.keys())
    defined_methods += [base.__dict__.keys() for base in bases]

    # Checking for missing methods
    missing_methods = [meth for meth in cls.needed_methods
                       if meth not in defined_methods]

    # Raising if there are unexpected missing methods
    if missing_methods and name != "Modifier":
      raise DefinitionError(
        f'Class {name} is missing the required method(s): '
        f'{", ".join(missing_methods)}')

    # Otherwise, saving the class
    if name != "Modifier":
      cls.classes[name] = cls


class Modifier:
  """The base class for all modifier classes, simply allowing to keep track of
  them."""

  __metaclass__ = MetaModifier



# coding: utf-8

"""
Demonstration of a DIC controlled test.

This program is intended as a demonstration and is fully virtual. The strain is
measured using DIC and a PID controller drives the machine to apply a ramp of
strain.

No hardware required.
Requires the cv2 module to be installed.
"""

import crappy
import numpy as np
import cv2


class Apply_strain_img:
  """This class reshapes an image depending on input strain values."""

  def __init__(self,
               image: np.ndarray) -> None:
    """Sets the base image and initializes the necessary objects to use."""

    self._img = image

    # Building the lookup arrays for the cv2.remap method
    height, width, *_ = image.shape
    orig_x, orig_y = np.meshgrid(range(width), range(height))
    # These arrays correspond to the original state of the image
    self._orig_x = orig_x.astype(np.float32)
    self._orig_y = orig_y.astype(np.float32)

    # These arrays are meant to be added to the original image ones
    # If added as is, they correspond to a 100% strain state in both directions
    self._x_strain = self._orig_x * width / (width - 1) - width / 2
    self._y_strain = self._orig_y * height / (height - 1) - height / 2

  def __call__(self, exx: float, eyy: float) -> np.ndarray:
    """Returns the reshaped image, based on the given strain values."""

    exx /= 100
    eyy /= 100

    # The final lookup table is the sum of the original state ones plus the
    # 100% strain one weighted by a ratio
    transform_x = self._orig_x - (exx / (1 + exx)) * self._x_strain
    transform_y = self._orig_y - (eyy / (1 + eyy)) * self._y_strain

    return cv2.remap(self._img, transform_x, transform_y, 1)


if __name__ == "__main__":
  img = crappy.resources.speckle

  speed = .05  # Strain rate (%/s)

  # Simply create a sawtooth signal at 0.05 units/s between 0 and 1
  # Note that this generator takes no feedback
  generator = crappy.blocks.Generator(path=[
    {'type': 'cyclic_ramp', 'speed1': speed, 'condition1': 'target_Exx(%)>1',
     'speed2': -speed, 'condition2': 'target_Exx(%)<0', 'cycles': 100,
     'init_value': 0}],
     cmd_label='target_Exx(%)')

  crappy.link(generator, generator)

  # Our fake machine
  machine = crappy.blocks.Fake_machine(max_strain=1.7, k=5000, l0=20,
                                       plastic_law=lambda exx: 0,
                                       sigma={'F(N)': 0.5},
                                       cmd_label='pid')

  # The block performing the DIC
  dis = crappy.blocks.DISCorrel('', display_images=True,
                                labels=['t(s)', 'x', 'y', 'measured_Exx(%)',
                                        'measured_Eyy(%)'],
                                verbose=True, iterations=0, finest_scale=2,
                                image_generator=Apply_strain_img(img))

  # This modifier will generate an image with the values of strain
  # coming from the Fake_machine block
  crappy.link(machine, dis)

  # The PID block takes TWO inputs: the setpoint and the feedback
  # Here the setpoint is coming from the generator (target_Exx(%))
  # and the feedback is the strain measured with the DIC (measured_Exx(%))
  # The output is the command sent to the machine
  # In a real-world scenario, consider using out_min and out_max
  # to clamp the output value and i_limit to prevent over_integration.
  pid = crappy.blocks.PID(kp=0.5, ki=2, kd=0.05, target_label='target_Exx(%)',
                          input_label='measured_Exx(%)', send_terms=True)

  # We link the two inputs and the output
  # The order does not matter, the inputs are identified thanks to the
  # keywords target_label (setpoint) and input_label (feedback)
  crappy.link(generator, pid)
  crappy.link(dis, pid)
  crappy.link(pid, machine)

  # To see what is sent to the machine
  # Since send_terms is given to the PID block, it also returns the 3
  # channels of the PID (P, I and D). This can be useful when adjusting
  # the gains
  graph_pid = crappy.blocks.Grapher(('t(s)', 'pid'),
                                    ('t(s)', 'p_term'), ('t(s)', 'i_term'),
                                    ('t(s)', 'd_term'))
  crappy.link(pid, graph_pid)

  # To see the commanded and the measured strains
  graph_def2 = crappy.blocks.Grapher(('t(s)', 'measured_Exx(%)'),
                                     ('t(s)', 'target_Exx(%)'))
  crappy.link(dis, graph_def2)
  crappy.link(generator, graph_def2)

  crappy.start()



# coding: utf-8

from time import time
from typing import Optional, List
from .inout import InOut
from ..tool import ft232h_server as ft232h, Usb_server
from .._global import OptionalModule

try:
  from smbus2 import SMBus
except (ModuleNotFoundError, ImportError):
  SMBus = OptionalModule("smbus2")

try:
  import board
except (ModuleNotFoundError, ImportError):
  board = OptionalModule('board', 'Blinka is necessary to use the I2C bus')

try:
  import busio
except (ModuleNotFoundError, ImportError):
  busio = OptionalModule('busio', 'Blinka is necessary to use the I2C bus')

try:
  from adafruit_bus_device.i2c_device import I2CDevice
except (ModuleNotFoundError, ImportError):
  I2CDevice = OptionalModule('adafruit_bus_device',
                             'Blinka is necessary to use the I2C bus')

try:
  from adafruit_mcp9600 import MCP9600
except (ModuleNotFoundError, ImportError):
  MCP9600 = OptionalModule('adafruit_mcp9600',
                           'Blinka is necessary to use the I2C bus')

Mcp9600_registers = {'Hot Junction Temperature': 0x00,
                     'Junction Temperature Delta': 0x01,
                     'Cold Junction Temperature': 0x02,
                     'Raw Data ADC': 0x03,
                     'Status': 0x04,
                     'Thermocouple Sensor Configuration': 0x05,
                     'Device Configuration': 0x06,
                     'Alert 1 Configuration': 0x08,
                     'Alert 2 Configuration': 0x09,
                     'Alert 3 Configuration': 0x0A,
                     'Alert 4 Configuration': 0x0B,
                     'Alert 1 Hysteresis': 0x0C,
                     'Alert 2 Hysteresis': 0x0D,
                     'Alert 3 Hysteresis': 0x0E,
                     'Alert 4 Hysteresis': 0x0F,
                     'Alert 1 Limit': 0x10,
                     'Alert 2 Limit': 0x11,
                     'Alert 3 Limit': 0x12,
                     'Alert 4 Limit': 0x13,
                     'Device ID/REV': 0x20}

Mcp9600_filter_coefficients = {0: 0b000,
                               1: 0b001,
                               2: 0b010,
                               3: 0b011,
                               4: 0b100,
                               5: 0b101,
                               6: 0b110,
                               7: 0b111}

Mcp9600_thermocouple_types = {'J': 0b0000000,
                              'K': 0b0010000,
                              'T': 0b0100000,
                              'N': 0b0110000,
                              'S': 0b1000000,
                              'E': 0b1010000,
                              'B': 0b1100000,
                              'R': 0b1110000}

Mcp9600_adc_resolutions = {18: 0b0000000,
                           16: 0b0100000,
                           14: 0b1000000,
                           12: 0b1100000}

Mcp9600_sensor_resolutions = {0.0625: 0b00000000,
                              0.25: 0b10000000}

Mcp9600_modes = ['Hot Junction Temperature',
                 'Junction Temperature Delta',
                 'Cold Junction Temperature',
                 'Raw Data ADC']

Mcp9600_backends = ['Pi4', 'ft232h', 'blinka']


class Mcp9600(Usb_server, InOut):
  """Class for controlling Adafruit's MCP9600 thermocouple reader.

  The Mcp9600 InOut block is meant for reading temperature from an MCP9600
  board, using the I2C protocol. The output is in `°C`, except for one
  operating mode that returns Volts.
  """

  def __init__(self,
               backend: str,
               thermocouple_type: str,
               i2c_port: int = 1,
               device_address: int = 0x67,
               adc_resolution: int = 18,
               sensor_resolution: float = 0.0625,
               filter_coefficient: int = 0,
               mode: str = 'Hot Junction Temperature',
               ft232h_ser_num: Optional[str] = None) -> None:
    """Checks arguments validity.

    Args:
      backend (:obj:`str`): The backend for communicating with the MCP9600.
        Should be one of:
        ::

          'Pi4', 'ft232h', 'blinka'

        The `'Pi4'` backend is optimized but only works on boards supporting
        the :mod:`smbus2` module, like the Raspberry Pis. The `'blinka'`
        backend may be less performant and requires installing Adafruit's
        modules, but these modules are compatible with and maintained on a wide
        variety of boards. The `'ft232h'` backend allows controlling the
        ADS1115 from a PC using Adafruit's FT232H USB to I2C adapter. See
        :ref:`Crappy for embedded hardware` for details.
      thermocouple_type (:obj:`str`): The type of thermocouple plugged in the
        MCP9600. The available types are:
        ::

          'J', 'K', 'T', 'N', 'S', 'E', 'B', 'R'

      i2c_port(:obj:`int`, optional): The I2C port over which the MCP9600
        should communicate. On most Raspberry Pi models the default I2C port is
        `1`.
      device_address(:obj:`int`, optional): The I2C address of the MCP9600. The
        default address is `0x67`, but it is possible to change this setting
        using a specific setup involving the `ADDR` pin.
      adc_resolution(:obj:`int`, optional): The number of bits the ADC output
        is encoded on. The greater the resolution, the lower the sample rate.
        The available resolutions are:
        ::

          12, 14, 16, 18

      sensor_resolution(:obj:`float`, optional): The temperature measurement
        resolution in `°C`. It should be either `0.0625` or `0.25`. Setting the
        resolution to `0.25` will increase the sample rate, but the output
        temperature will be encoded on two bits less.
      filter_coefficient(:obj:`int`, optional): The MCP9600 features an
        integrated filter (see its documentation for the exact filter formula).
        When set to `0`, the filter is inactive. It is maximal when set to `7`.
        When active, the filter will prohibit fast temperature changes, thus
        limiting noise and smoothening the signal.
      mode(:obj:`str`, optional): Four different values can be accessed when
        measuring a temperature: the temperature of the thermocouple (hot
        junction temperature), the temperature of the MCP9600 board (cold
        junction temperature), the temperature calculated from the ADC data and
        thermocouple type but not yet cold junction-compensated (junction
        temperature delta), and the raw ADC measurement of the voltage
        difference in the thermocouple (raw data ADC, in Volts). The available
        modes are thus:
        ::

          'Hot Junction Temperature',
          'Junction Temperature Delta',
          'Cold Junction Temperature',
          'Raw Data ADC'

      ft232h_ser_num (:obj:`str`, optional): If backend is `'ft232h'`, the
        serial number of the ft232h to use for communication.

    """

    if not isinstance(backend, str) or backend not in Mcp9600_backends:
      raise ValueError("backend should be in {}".format(Mcp9600_backends))
    self._backend = backend

    Usb_server.__init__(self,
                        serial_nr=ft232h_ser_num if ft232h_ser_num else '',
                        backend=backend)
    InOut.__init__(self)
    current_file, block_number, command_file, answer_file, block_lock, \
        current_lock = super().start_server()

    if backend == 'Pi4':
      self._bus = SMBus(i2c_port)
    elif backend == 'ft232h':
      self._bus = ft232h(mode='I2C',
                         block_number=block_number,
                         current_file=current_file,
                         command_file=command_file,
                         answer_file=answer_file,
                         block_lock=block_lock,
                         current_lock=current_lock,
                         serial_nr=ft232h_ser_num,
                         i2c_speed=20E3)
    elif backend == 'blinka':
      self._i2c = busio.I2C(board.SCL, board.SDA, frequency=200000)

    self._device_address = device_address

    if sensor_resolution not in Mcp9600_sensor_resolutions:
      raise ValueError("sensor_resolution should be in {}".format(list(
        Mcp9600_sensor_resolutions.keys())))
    else:
      self._sensor_resolution = sensor_resolution

    if adc_resolution not in Mcp9600_adc_resolutions:
      raise ValueError("adc_resolution should be in {}".format(list(
        Mcp9600_adc_resolutions.keys())))
    elif adc_resolution != 18 and backend == 'blinka':
      raise ValueError("It is not possible to set adc_resolution using the"
                       "backend blinka.")
    else:
      self._adc_resolution = adc_resolution

    if thermocouple_type not in Mcp9600_thermocouple_types:
      raise ValueError("thermocouple_type should be in {}".format(list(
        Mcp9600_thermocouple_types.keys())))
    else:
      self._thermocouple_type = thermocouple_type

    if filter_coefficient not in Mcp9600_filter_coefficients:
      raise ValueError("filter_coefficient should be an integer between "
                       "0 and 7")
    else:
      self._filter_coefficient = filter_coefficient

    if mode not in Mcp9600_modes:
      raise ValueError("mode should be in {}".format(Mcp9600_modes))
    elif mode == 'Raw Data ADC' and backend == 'blinka':
      raise ValueError('The Raw Data ADC mode is not available using the '
                       'backend blinka.')
    else:
      self._mode = mode

  def open(self) -> None:
    """Sets the I2C communication and device."""

    if self._backend == 'blinka':
      self._mcp = MCP9600(self._i2c,
                          address=self._device_address,
                          tctype=self._thermocouple_type,
                          tcfilter=self._filter_coefficient)
      self._mcp.ambient_resolution = self._sensor_resolution == 0.25

    else:
      if not self._is_connected():
        raise IOError("The MCP9600 is not connected")

      # Setting the sensor according to the user parameters
      config_sensor = Mcp9600_filter_coefficients[self._filter_coefficient]
      config_sensor |= Mcp9600_thermocouple_types[self._thermocouple_type]

      self._bus.write_i2c_block_data(self._device_address, Mcp9600_registers[
          'Thermocouple Sensor Configuration'], [config_sensor])

      # Setting the device according to the user parameters
      config_device = Mcp9600_sensor_resolutions[self._sensor_resolution]
      config_device |= Mcp9600_adc_resolutions[self._adc_resolution]
      config_device |= 0b00  # Normal operating mode
      self._bus.write_i2c_block_data(self._device_address,
                                     Mcp9600_registers['Device Configuration'],
                                     [config_device])

  def get_data(self) -> List[float]:
    """Reads the registers containing the conversion result.

    The output is in `°C` for all modes except the raw data ADC one, which
    outputs Volts.

    Returns:
      :obj:`list`: A list containing the timeframe and the output value
    """

    if self._backend == 'blinka':
      out = [time()]
      if self._mode == 'Hot Junction Temperature':
        out.append(self._mcp.temperature)
      elif self._mode == 'Junction Temperature Delta':
        out.append(self._mcp.delta_temperature)
      elif self._mode == 'Cold Junction Temperature':
        out.append(self._mcp.ambient_temperature)

    else:
      # Starting a conversion
      value = self._bus.read_i2c_block_data(self._device_address,
                                            Mcp9600_registers['Status'], 1)[0]
      self._bus.write_i2c_block_data(self._device_address,
                                     Mcp9600_registers['Status'],
                                     [value & 0xBF])

      # Waiting for the conversion to complete
      t0 = time()
      while not self._data_available():
        if time() - t0 > 1:
          raise TimeoutError('Waited too long for data to be ready !')

      out = [time()]

      # The number of output bits varies according to the selected mode
      # Temperature-modes outputs are 2 bytes long
      if self._mode in ['Hot Junction Temperature',
                        'Junction Temperature Delta',
                        'Cold Junction Temperature']:
        block = self._bus.read_i2c_block_data(self._device_address,
                                              Mcp9600_registers[self._mode], 2)
        value_raw = ((block[0] << 8) | block[1])

        # Converting the raw output value into °C
        value = value_raw / 16
        if block[0] >> 7:
          value -= 4096
        out.append(value)

      # Raw ADC output is 3 bytes long
      else:
        block = self._bus.read_i2c_block_data(self._device_address,
                                              Mcp9600_registers[self._mode], 3)
        value_raw = (block[0] << 16) | (block[1] << 8) | block[2]

        # Converting the raw output value into Volts
        # The 2µV resolution is given in the MCP9600 datasheet
        if block[0] >> 1 & 1:
          value_raw -= 2 ** 18

        out.append(value_raw * 2E-6)

    return out

  def close(self) -> None:
    """Switches the MCP9600 to shutdown mode and closes the I2C bus.."""

    if self._backend != 'blinka':
      # switching to shutdown mode, keeping configuration
      value = self._bus.read_i2c_block_data(self._device_address,
                                            Mcp9600_registers[
                                             'Device Configuration'], 1)[0]
      value &= 0xFD
      value |= 0x01
      self._bus.write_i2c_block_data(self._device_address,
                                     Mcp9600_registers['Device Configuration'],
                                     [value])
      self._bus.close()

  def _data_available(self) -> int:
    """Reads the data available bit.

    Returns: `1` if data is available, else `0`
    """

    status = self._bus.read_i2c_block_data(self._device_address,
                                           Mcp9600_registers['Status'], 1)[0]

    # The MCP9600 features an over-temperature protection
    if status & 0x10:
      raise ValueError("Too hot for selected thermocouple !")

    return status & 0x40

  def _is_connected(self) -> bool:
    """Tries reading a byte from the device.

    Returns: :obj:`True` if reading was successful, else :obj:`False`
    """

    try:
      self._bus.read_byte(self._device_address)
      return True
    except IOError:
      return False



# coding: utf-8

from numpy import interp, array, searchsorted, concatenate
from itertools import chain

from .block import Block


class Multiplex(Block):
  """This block takes data from upstream blocks as input and interpolates it to
  output all labels in a common time basis.

  It is useful for synchronizing data acquired from different sensors, e.g. to
  plot a real-time stress-strain curve. This block is however quite
  resource-consuming, so it is preferable to perform interpolation in
  post-processing if real-time is not needed.

  Note:
    This block doesn't truly output data in real-time as it needs to wait for
    data from all the upstream blocks before performing the interpolation.
    So it should only be used with care when it is an input of a
    decision-making block. This is especially true when the upstream blocks
    have very different sample rates.
  """

  def __init__(self,
               time_label: str = 't(s)',
               freq: float = 200,
               verbose: bool = False) -> None:
    """Sets the args and initializes the parent class.

    Args:
      time_label: The label carrying the time information.
      freq : The sample rate for the interpolation, and the target looping
        frequency for the block. If this value is set too high and your machine
        cannot keep up, the block will most likely lag.
      verbose: If :obj:`True`, prints information about the looping frequency
        of the block.
    """

    Block.__init__(self)

    # Initializing the attributes
    self._time_label = time_label
    self.freq = freq
    self.verbose = verbose
    self._t = 0
    self._dt = 1 / self.freq

    # Creating the different dicts holding information
    self._values = dict()
    self._timestamps = dict()
    self._labels_to_get = dict()
    self._label_to_link = dict()

  def begin(self) -> None:
    """Receiving the first data from the upstream blocks, and checking that it
    is valid.

    If part of the data is not valid, warning the user. For the valid data,
    initializing the different dicts with it.
    """

    for link in self.inputs:
      # First, receiving data from each incoming link
      data = link.recv_chunk()

      # Handling the case when the time label is absent from the data
      if self._time_label not in data:
        print(f'WARNING : Cannot multiplex data coming from link {link.name} '
              f'as it does not contain the time label ({self._time_label})')
        continue

      # Extracting the timestamps from the received data
      timestamp = data.pop(self._time_label)

      # Determining which labels to use for multiplexing and warning the user
      # if two similar labels found
      self._labels_to_get[link] = [label for label in data if label not in
                                   chain(*self._labels_to_get.values())]
      if len(self._labels_to_get[link]) != len(data):
        print(f"WARNING : Got identical label(s)" 
              f"""{tuple(label for label in data if label 
                         not in self._labels_to_get[link])}""" 
              f"from at least two different links, on of them won't be used "
              f"for multiplexing.")

      self._label_to_link.update({label: link for label in
                                  self._labels_to_get[link]})

      # Storing the timestamps for the link
      self._timestamps[link] = array(timestamp)
      # Storing the values for the labels that were kept
      self._values.update({label: array(values) for label, values in
                           data.items() if label in self._labels_to_get[link]})

  def loop(self) -> None:
    """Receives data, interpolates it, and sends it to the downstream
    blocks."""

    self._get_data()
    self._send_data()

  def finish(self) -> None:
    """Just sending any remaining data."""

    self._send_data()

  def _get_data(self) -> None:
    """Receives data from the upstream links."""

    for link in self.inputs:
      # Receiving data from each link, non-blocking to prevent accumulation
      data = link.recv_chunk(blocking=False)
      # Processing only the valid labels
      if data is not None and link in self._labels_to_get:
        # Saving the timestamps
        self._timestamps[link] = concatenate((self._timestamps[link],
                                              array(data[self._time_label])))
        # Saving the other values
        for label in self._labels_to_get[link]:
          self._values[label] = concatenate((self._values[label],
                                             array(data[label])))

  def _send_data(self) -> None:
    """Interpolates the previously received data, and sends the result to the
    downstream blocks."""

    # Making sure all the necessary data has been received for interpolating
    if all(timestamp[-1] > self._t for timestamp
           in self._timestamps.values()) and any(
      timestamp.size for timestamp in self._timestamps.values()):

      # Getting the maximum timestamp common to all the labels
      max_t = min(times[-1] for times in self._timestamps.values())
      # Deducing the number of time intervals to interpolate on
      n_samples = int((max_t - self._t) // self._dt) + 1
      # Creating the array of timestamps for interpolation
      new_times = [self._t + i * self._dt for i in range(n_samples)]

      # For each link, getting the index for trimming data after interpolation
      last_indexes = {link: searchsorted(self._timestamps[link], new_times[-1],
                                         side='right')
                      for link in self.inputs}

      # This dict stores the values to send
      to_send = {self._time_label: new_times}

      # For each label, interpolating the data
      for label in self._values:
        link = self._label_to_link[label]
        to_send[label] = list(interp(new_times, self._timestamps[link],
                                     self._values[label]))

        # Trimming the values to save some memory
        self._values[label] = self._values[label][last_indexes[link]:]

      # Also trimming the timestamps to save some memory
      for link in self.inputs:
        self._timestamps[link] = self._timestamps[link][last_indexes[link]:]

      # Updating the current time value
      self._t = new_times[-1] + self._dt

      # Finally, sending the data to downstream blocks
      while any(to_send.values()):
        self.send({label: values.pop(0) for label, values in to_send.items()})



# coding: utf-8

from typing import Union, Dict
from .path import Path, condition_type


class Protection(Path):
  """Depending on two different conditions checked at each loop, this path can
  output one between 3 constant values.

  It is especially useful for controlling processes that need to behave
  differently based on given conditions, e.g. for preventing a heating element
  from overheating or a motor from driving too far.
  """

  def __init__(self,
               _last_time: float,
               _last_cmd: float,
               condition1: Union[str, condition_type],
               condition2: Union[str, condition_type],
               value1: float,
               value2: float,
               value0: float = 0) -> None:
    """Sets the args and initializes the parent class.

    Args:
      _last_time: The last timestamp when a command was generated. For internal
        use only, do not overwrite.
      _last_cmd: The last sent command. For internal use only, do not
        overwrite.
      condition1: The first condition checked by the path. Refer to
        :ref:`generator path` for more info.
      condition2: The second condition checked by the path. Refer to
        :ref:`generator path` for more info.
      value1: Value to send when ``condition1`` is met.
      value2: Value to send when ``condition2`` is met and ``condition1`` is
        not met.
      value0: Value to send when neither ``condition1`` nor ``condition2`` are
        met.

    Note:
      This generator path never ends, it doesn't have a stop condition.
    """

    Path.__init__(self, _last_time, _last_cmd)

    # Setting the attributes
    self._value0 = value0
    self._value1 = value1
    self._value2 = value2
    self._condition1 = self.parse_condition(condition1)
    self._condition2 = self.parse_condition(condition2)
    self._prev = self._value0

  def get_cmd(self, data: Dict[str, list]) -> float:
    """Sends either value1 if condition1 is met, or value2 if only condition2
    is met, or value0 if none of the conditions are met."""

    # Case when data has been received
    if any(data.values()):

      # Send value1 if the first condition is met
      if self._condition1(data):
        self._prev = self._value1
        return self._value1

      # Send value2 if only the second condition is met
      elif self._condition2(data):
        self._prev = self._value2
        return self._value2

      # Send value0 if no condition is met
      else:
        self._prev = self._value0
        return self._value0

    # If no data received, return the last sent value
    else:
      return self._prev



# coding: utf-8

"""
Code demonstrating the use of a linear actuator to follow the videoextensometry
markers during a test with large strains.
"""

import crappy

ve = crappy.blocks.Video_extenso(camera='Ximea_cv', show_image=True)

ad = crappy.blocks.AutoDrive(
    actuator={'name': 'CM_drive', 'port': '/dev/ttyUSB0'}, direction='X-')

graph_extenso = crappy.blocks.Grapher(('t(s)', 'Exx(%)'), ('t(s)', 'Eyy(%)'))

crappy.link(ve, graph_extenso)

crappy.link(ve, ad)

crappy.start()



# coding: utf-8

from .modifier import Modifier
from typing import Optional, Dict, Any


class Integrate(Modifier):
  """This modifier integrates the data of a label over time and adds the
  integration value to the returned data."""

  def __init__(self,
               label: str,
               time_label: str = 't(s)',
               out_label: Optional[str] = None) -> None:
    """Sets the args and initializes the parent class.

    Args:
      label: The label whose data to integrate over time.
      time_label: The label carrying the time information.
      out_label: The label carrying the integration value. If not given,
        defaults to ``'i_<label>'``.
    """

    super().__init__()
    self._label = label
    self._time_label = time_label
    self._out_label = out_label if out_label is not None else f'i_{label}'

    self._last_t = None
    self._last_val = None
    self._integration = 0

  def evaluate(self, data: Dict[str, Any]) -> Dict[str, Any]:
    """Gets the data from the upstream block, updates the integration value and
    returns it."""

    # For the first received data, storing it and returning 0
    if self._last_t is None or self._last_val is None:
      self._last_t = data[self._time_label]
      self._last_val = data[self._label]
      data[self._out_label] = self._integration
      return data

    # Updating the integrated value with the latest received values
    t = data[self._time_label]
    val = data[self._label]
    self._integration += (t - self._last_t) * (val + self._last_val) / 2
    # Updating the stored data
    self._last_t = t
    self._last_val = val

    # Returning the updated data
    data[self._out_label] = self._integration
    return data



# coding: utf-8

import numpy as np
from typing import Dict, Any, Optional

from .modifier import Modifier


class Mean(Modifier):
  """Modifier waiting for a given number of data points to be received, then
  returning their average, and starting all over again.

  Unlike :ref:`Moving average`, it only returns a value once every ``n_points``
  points.
  """

  def __init__(self, n_points: int = 100) -> None:
    """Sets the args and initializes the parent class.

    Args:
      n_points: The number of points on which to compute the average.
    """

    super().__init__()
    self._n_points = n_points
    self._buf = None

  def evaluate(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Receives data from the upstream block, and computes the average of every
    label once the right number of points have been received. Then empties the
    buffer and returns the averages.

    If there are not enough points, doesn't return anything.
    """

    # Initializing the buffer
    if self._buf is None:
      self._buf = {key: [value] for key, value in data.items()}

    ret = {}
    for label in data:
      # Updating the buffer with the newest data
      self._buf[label].append(data[label])

      # Once there's enough data in the buffer, calculating the average value
      if len(self._buf[label]) == self._n_points:
        try:
          ret[label] = np.mean(self._buf[label])
        except TypeError:
          ret[label] = self._buf[label][-1]

        # Resetting the buffer
        self._buf[label].clear()

    if ret:
      return ret



# coding: utf-8

from time import sleep
from typing import List

from .inout import InOut
from .._global import OptionalModule

try:
  from serial import Serial
  from serial.serialutil import SerialException
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("serial", "Please install the module serial to use "
                          "the Gsm InOut: pip install pyserial")


class Gsm(InOut):
  """Block for sending messages by SMS to given phone numbers.

  Important:
    This block should be associated with a modifier to manage the messages
    to send.
  """

  def __init__(self,
               numbers: List[str],
               port: str = "/dev/ttyUSB0",
               baudrate: int = 115200) -> None:
    """Checks arguments validity.

    Args:
      numbers(:obj:`list`): The list of numbers the messages will be sent to.
        The syntax is the following :
        ::

          ["0611223344"]

      port (:obj:`str`, optional): Serial port the GSM is connected to.
      baudrate(:obj:`int`, optional): Serial baudrate, between 1200 and 115200.
    """

    super().__init__()
    try:
      self._ser = Serial(port, baudrate)
    except SerialException:
      raise SerialException("GSM not connected or wrong port")

    # Change the type of numbers to bytes rather than string
    self._numbers = [number.encode('utf-8') for number in numbers]

  def open(self) -> None:
    """Sends ``"AT"`` to the GSM and waits for the response : ``"OK"``. """

    self._ser.write(b'AT' + b'\r\n')
    count = 0
    while count <= 2:
      sleep(0.1)
      data = ""
      while self._ser.inWaiting() > 0:
        data += self._ser.read(self._ser.inWaiting()).decode()
      if "OK" in data:
        return
      count += 1
    raise TimeoutError("GSM is not responding")

  def set_cmd(self, *cmd: str) -> None:
    """Sends an SMS whose text is the :obj:`str` received as command to all the
    phone numbers.

    Doesn't send anything if the string is empty, and raises a :exc:`TypeError`
    if the command is not a :obj:`str`.
    """

    if not isinstance(cmd[0], str):
      raise TypeError("Message should be a string")
    if cmd[0] != "":
      self._send_mess(cmd[0])

  def _send_mess(self, message: str) -> None:
    """Commands the GSM to send a message to all the phone numbers.

    Args:
      message: The text message to send.
    """

    for number in self._numbers:
      count = 0
      self._ser.write(b'AT' + b'\r\n')
      w_buff = [b"AT+CMGF=1\r\n",
                b"AT+CMGS=\"" + number + b"\"\r\n", message.encode()]
      while count <= 2:
        data = ""
        while self._ser.inWaiting() > 0:
          data += self._ser.read(self._ser.inWaiting()).decode()
          # Get all the answers in Waiting
        if data:
          if count < 2:
            sleep(1)
            self._ser.write(w_buff[count])
            # Put the message in text mode then enter the
            # number to contact
          if count == 2:
            sleep(0.5)
            self._ser.write(w_buff[2])  # Write the message
            self._ser.write(b"\x1a\r\n")
            # 0x1a : send   0x1b : Cancel send
          count += 1

  def close(self) -> None:
    """Closes the serial port."""

    self._ser.close()



# coding: utf-8

"""
Using directly the camera object to take a picture.

Use `CONFIG=True` to open the configuration GUI.

Required hardware:
  - Any camera (replace Webcam by the desired camera if necessary)
"""

import SimpleITK as Sitk

import crappy

if __name__ == "__main__":
  CONFIG = False  # True

  cam = crappy.camera.Webcam()
  cam.open()
  if CONFIG:
    crappy.tool.Camera_config(cam).main()
  Sitk.WriteImage(Sitk.GetImageFromArray(cam.get_image()[1]), "photo.tiff")
  cam.close()



# coding:utf-8

import Tix
from Tkinter import *
from crappy.actuator import actuator_list


class Interface(Frame):
  """Creates a graphic interface that allows to connect the actuator via a
  serial connection and to send command in terms of speed or position."""

  def __init__(self, root, **kwargs):
    Frame.__init__(self, root, width=1000, height=1000, **kwargs)
    root.geometry("395x320")
    root.title("Pilotage")

    # Création des onglets
    monnotebook = Tix.NoteBook(root)
    monnotebook.add("page1", label="Configuration")
    monnotebook.add("page2", label="Velocity Mode")
    monnotebook.add("page3", label="Motion Task")

    p1 = monnotebook.subwidget_list["page1"]
    p2 = monnotebook.subwidget_list["page2"]
    p3 = monnotebook.subwidget_list["page3"]

    fra1 = Canvas(p1)
    fra2 = Canvas(p2)
    fra3 = Canvas(p3)

    fra1.pack(expand=1, fill=Tix.BOTH)
    fra2.pack(expand=1, fill=Tix.BOTH)
    fra3.pack(expand=1, fill=Tix.BOTH)
    # fin création des onglets

    # Début onglet 1
    # create a frame in canvas(p1)
    sous_fra = Frame(fra1, width=400, borderwidth=2, relief=GROOVE)
    self.portLabel = Label(sous_fra, text="Serial ser:")  # create a label
    # create a variable, it will contain ser selection
    # self.myPortCombo = Tix.StringVar()
    self.myPortCombo = Tix.StringVar(sous_fra, "/dev/ttyUSB0")

    self.actuatorNameCombo = Tix.StringVar()
    self.actuatorNameCombo.set("CM_drive")
    # create a combobox, it will contain names of ports
    self.actuatorCombo = Tix.ComboBox(sous_fra, editable=1, dropdown=1,
                                      variable=self.actuatorNameCombo)
    # configure the combobox in read only
    self.actuatorCombo.entry.config(state='readonly')
    for m in actuator_list:
      self.actuatorCombo.insert(0, m)

    self.actuatorLabel = Label(sous_fra, text="Actuator name:")
    self.portCombo = Entry(sous_fra, textvariable=self.myPortCombo)
    self.baudrateLabel = Label(sous_fra, text="Baudrate:")  # create a label
    # create a variable, it will contain baudrate selection
    # self.baudCombo = Tix.StringVar()
    self.baudCombo = Tix.StringVar(sous_fra, "9600")
    self.baudrateCombo = Entry(sous_fra, textvariable=self.baudCombo)
    self.num_device = Tix.IntVar()
    # create a label
    self.num_device_label = Label(sous_fra, text="Device number:")
    self.num_device_entry = Spinbox(sous_fra, from_=1, to=10,
                                    textvariable=self.num_device)
    self.actuatorNameCombo.trace('w', self.callback)
    # create connect Button
    self.connection = Button(sous_fra, text="Connect", command=self.connection)
    sous_fra1 = Frame(fra1, width=400)  # create a second frame in canvas(p1)
    # create examine location button
    self.location = Button(sous_fra1, text="Examine location",
                           command=self.examine_location, width=10)
    self.location.config(state='disabled')  # disable the state of the button
    # create 'location:' label
    location_label = Label(sous_fra1, text="Location:")
    # create a variable, it will contain the result of examine_location method
    self.Resultat = StringVar()
    self.Resultat.set("0")  # set the variable to zero
    # create a label, it will show the variable
    resultat_label = Label(sous_fra1, textvariable=self.Resultat)
    self.resetZero = Button(sous_fra1, text=" Reset ",
                            command=self.reset_servostar)
    # create cancel button
    self.quit = Button(sous_fra1, text="Cancel", command=root.quit)
    # Fin onglet 1

    # Début onglet 2
    # create a frame in canvas(p2)
    sous_fra2 = Frame(fra2, width=400)
    # create a second frame in canvas(p2)
    sous_fra2bis = Frame(fra2, width=400)
    # create a variable, it will contain the value of speed_entry
    self.speedVar = StringVar()
    # create 'Velocity:' label
    speed_label = Label(sous_fra2bis, text="Velocity:")
    # create an entry, it will contain a velocity choose by the user
    speed_entry = Entry(sous_fra2bis, textvariable=self.speedVar, width=5)
    # create a advance button
    self.advanceButton = Button(sous_fra2bis, text="advance",
                                command=self.advance, width=10)
    # disable the state of the button
    self.advanceButton.config(state='disabled')
    # create recoil button
    self.recoilButton = Button(sous_fra2bis, text="recoil",
                               command=self.recoil, width=10)
    # disable the state of the button
    self.recoilButton.config(state='disabled')
    # create stop button
    self.stopVMButton = Button(sous_fra2bis, text="STOP",
                               command=self.stop_motion, width=10)
    self.stopVMButton.config(state='disabled')
    self.defineZeroButton = Button(sous_fra2bis, text="Define Zero",
                                   command=self.define_zero, width=10)
    self.defineZeroButton.config(state='disabled')
    # Début onglet 3
    sous_fra3_2 = Frame(fra3, width=400)  # create a frame in canvas(p3)
    self.moveZeroButton = Button(sous_fra3_2, text="Move Home",
                                 command=self.move_zero, width=15)
    # create move home button

    # disable the state of the button
    self.moveZeroButton.config(state='disabled')

    # create 'position:' label
    position_label = Label(sous_fra3_2, text="position :")
    # create a variable, it will contain the value of entry2
    self.position = StringVar()
    # create an entry, it will contain the positon choose by the user
    self.entry2 = Entry(sous_fra3_2, textvariable=self.position)
    # pick out the widget that will receive keyboard events
    self.entry2.focus_set()

    speed_label = Label(sous_fra3_2, text="speed:")  # create 'position:' label
    # create a variable, it will contain the value of entry2
    self.speed = StringVar()
    # create an entry, it will contain the positon choose by the user
    self.entry3 = Entry(sous_fra3_2, textvariable=self.speed)
    # pick out the widget that will receive keyboard events
    self.entry3.focus_set()

    # create 'motionType:' label
    motion_type_label = Label(sous_fra3_2, text="motion type :")
    # create a variable, it will contain the value of entry4
    self.motionType = StringVar()

    sous_fra3 = Frame(fra3, width=400)  # create a frame in canvas(p3)
    sous_sous_fra3 = Frame(sous_fra3, width=400)  # create a frame in sous_fra3
    # create move button
    self.moveButton = Button(sous_sous_fra3, text="Move", command=self.move)
    self.moveButton.config(state='disabled')  # disable the state of the button
    # create STOP button
    self.stopMTButton = Button(sous_sous_fra3, text="STOP",
                               command=self.stop_motion)
    self.stopMTButton.config(state='disabled')
    # create a second frame in sous_fra3
    sous_sous_fra3bis = Frame(sous_fra3, width=400)
    # placement des widgets onglet 1
    # show widgets on canvas(p1)
    sous_fra.grid(padx=10, pady=10)  #
    self.portLabel.grid(row=1, column=0, sticky="sw", padx=10, pady=10)
    self.portCombo.grid(row=1, column=1, sticky="sw", padx=10, pady=10)
    self.baudrateLabel.grid(row=2, column=0, sticky="sw", padx=10, pady=10)
    self.baudrateCombo.grid(row=2, column=1, sticky="sw", padx=10, pady=10)
    self.actuatorLabel.grid(row=3, column=0, sticky="sw", padx=10, pady=10)
    self.actuatorCombo.grid(row=3, column=1, sticky="sw", padx=10, pady=10)
    self.connection.grid(row=5, column=1, sticky="se")
    # placement widget sous frame onglet 1
    # show widgets on frame canvas(p1)
    sous_fra1.grid()
    # self.enable.grid(row=1, column=0,sticky= "sw", padx=10,pady=10)
    # self.disable.grid(row=2, column=0,sticky= "sw", padx=10,pady=10)
    # self.resetZero.grid(row=3, column=0,sticky= "sw", padx=10,pady=10)
    self.location.grid(row=5, column=0, sticky="s", padx=10, pady=10)
    location_label.grid(row=5, column=1)
    resultat_label.grid(row=5, column=2)
    self.resetZero.grid(row=5, column=4)
    # self.quit.grid(row=4, column=4, sticky= "e")

    # placement des widgets onglet 2
    # show widgets on canvas(p2)
    sous_fra2.grid(row=0, column=0, padx=10, pady=10)
    sous_fra2bis.grid(row=1, column=0, padx=10, pady=10)
    # self.init_modeButton.grid(row=0, column=0, padx=10,pady=10)
    speed_label.grid(row=0, column=0, sticky='w')
    speed_entry.grid(row=0, column=2, sticky='w')
    self.recoilButton.grid(row=0, column=3, sticky='w')
    self.advanceButton.grid(row=1, column=3, sticky='w')
    self.stopVMButton.grid(row=2, column=3, sticky='w')
    self.defineZeroButton.grid(row=3, column=3, sticky='w')

    # placement des widgets onglet 3
    #  show widgets on canvas(p3)
    sous_fra3_2.grid(padx=10, pady=10)
    self.moveZeroButton.grid(row=0, column=0, padx=10, pady=10)
    position_label.grid(row=1, column=0, sticky='w')
    speed_label.grid(row=2, column=0, sticky='w')
    motion_type_label.grid(row=3, column=0, sticky='w')
    self.entry2.grid(row=1, column=1, sticky='w')
    self.entry3.grid(row=2, column=1, sticky='w')
    Radiobutton(sous_fra3_2, text="absolute", variable=self.motionType,
                value=True).grid()
    Radiobutton(sous_fra3_2, text="relative", variable=self.motionType,
                value=False).grid()
    sous_fra3.grid(row=3, column=0)
    sous_sous_fra3.grid(row=0, column=0)
    sous_sous_fra3bis.grid(row=1, column=0)
    self.moveButton.grid(row=0, column=1)
    self.stopMTButton.grid(row=0, column=2)
    # show notebooks
    monnotebook.pack(side=LEFT, fill=Tix.BOTH, expand=1, padx=5, pady=5)

  # function to initialize the connection
  def connection(self):
    if "" in [self.myPortCombo.get(), self.baudCombo.get(),
              self.actuatorNameCombo.get()]:
      print('you must choose actuator name and configuration.')
    else:
      try:
        actuator = actuator_list[self.actuatorNameCombo.get()]
        if self.actuatorNameCombo.get().capitalize() == "Oriental":
          self.actuator = actuator(port=self.myPortCombo.get(),
                                   num_device=int(self.num_device_entry.get()),
                                   baudrate=self.baudCombo.get())
        else:
          self.actuator = actuator(port=self.myPortCombo.get(),
                                   baudrate=self.baudCombo.get())
        self.actuatorName = self.actuatorNameCombo.get()
        print('connection')
        self.actuator.open()
        self.location.config(state='normal')
        self.moveButton.config(state='normal')
        self.recoilButton.config(state='normal')
        self.advanceButton.config(state='normal')
        self.moveZeroButton.config(state='normal')
        self.defineZeroButton.config(state='normal')
        self.stopMTButton.config(state='normal')
        self.stopVMButton.config(state='normal')

      except Exception as e:
        print(' Connection error:', e)

  def callback(self, *_):
    # print args
    # print 'test: ', self.actuatorNameCombo.get().capitalize()
    if self.actuatorNameCombo.get().capitalize() == "Oriental":
      self.num_device_label.grid(row=4, column=0, sticky="sw", padx=10,
                                 pady=10)
      self.num_device_entry.grid(row=4, column=1, sticky="sw", padx=10,
                                 pady=10)
    else:
      self.num_device_entry.grid_forget()
      self.num_device_label.grid_forget()

  def define_zero(self):
    if self.actuatorName == "CmDrive":
      self.actuator.ser.close()
      self.actuator.ser.open()
      self.actuator.ser.write('P=0\r')
      # self.actuator.ser.readline()
      self.actuator.ser.close()
    else:
      try:
        self.actuator.set_home()
      except NotImplementedError:
        print("Not implemented yet.")

  def reset_servostar(self):
    self.actuator.reset()

  # function to examine the location of the actuator
  def examine_location(self):
    location = self.actuator.get_position()
    self.Resultat.set(location)

  # function to move home
  def move_zero(self):
    self.actuator.move_home()
    print('moving home')

  # function to apply a motion task
  def move(self):
    if self.motionType.get() == "" or self.position.get() == "":
      print("one of the entry is empty")
    else:
      if self.actuatorName == "CmDrive":
        if self.motionType.get() == '1':
          self.actuator.set_position(int(self.position.get()), None,
                                     'absolute')
          print('MA mode')
        else:
          self.actuator.set_position(int(self.position.get()), None,
                                     'relative')
          print('MR mode')
      else:
        self.actuator.set_position(int(self.position.get()),
                                   int(self.speed.get()))

  # function to advance the actuator on velocity mode
  def advance(self):
    if self.speedVar.get() == "":
      print('choose velocity')
    else:
      self.actuator.set_speed(int(self.speedVar.get()))
      print('the actuator goes up with speed=%i' % int(self.speedVar.get()))

  # function to recoil the actuator on velocity mode
  def recoil(self):
    if self.speedVar.get() == "":
      print('choose velocity')
    else:
      self.actuator.set_speed(-int(self.speedVar.get()))
      print('the actuator goes down with speed=%i' % int(self.speedVar.get()))

  # function to stop a motion
  def stop_motion(self):
    self.actuator.stop()
    print('Motion has been stopped')


if __name__ == '__main__':
  root = Tix.Tk()
  interface = Interface(root)
  interface.mainloop()
  interface.destroy()



# coding: utf-8

"""
A fully autonomous example of a tensile test using a virtual machine.

This program is meant to replicate a tensile test without the need for an
actual machine.

No hardware required.
"""

import crappy

if __name__ == "__main__":
  speed = 5/60  # mm/sec

  generator = crappy.blocks.Generator(path=sum([[
    {'type': 'constant', 'value': speed,
     'condition': 'Exx(%)>{}'.format(i / 3)},
    {'type': 'constant', 'value': -speed, 'condition': 'F(N)<0'}]
    for i in range(5)], []), spam=False, cmd_label='cmd')

  machine = crappy.blocks.Fake_machine()

  crappy.link(generator, machine)
  crappy.link(machine, generator)

  # graph_def = crappy.blocks.Grapher(('t(s)', 'Exx(%)'))
  # crappy.link(machine, graph_def)

  graph_f = crappy.blocks.Grapher(('Exx(%)', 'F(N)'))
  crappy.link(machine, graph_f)

  # graph_x = crappy.blocks.Grapher(('t(s)', 'x(mm)'))
  # crappy.link(machine, graph_x)

  crappy.start()



# coding: utf-8

from typing import Optional
import tkinter as tk
from .cameraConfigBoxes import Camera_config_with_boxes
from .cameraConfigTools import Box, Spot_boxes


class DISVE_config(Camera_config_with_boxes):
  """Class similar to :ref:`Camera configuration` but also displaying the
  bounding boxes of the regions defined as patches.

  It is meant to be used for configuring the :ref:`Disve` block.
  """

  def __init__(self, camera, patches: Spot_boxes) -> None:
    """Sets the patches and initializes the parent class.

    Args:
      camera: The camera object in charge of acquiring the images.
      patches: The patches to follow for image correlation.
    """

    super().__init__(camera)

    # Setting the patches
    self._spots = patches

  def _on_img_resize(self, _: Optional[tk.Event] = None) -> None:
    """Same as in the parent class except it also draws the patches on top of
    the displayed image."""

    self._draw_spots()
    self._resize_img()
    self._display_img()
    self.update()

  def _update_img(self) -> None:
    """Same as in the parent class except it also draws the patches on top of
    the displayed image."""

    _, img = self._camera.get_image()

    self._cast_img(img)
    self._draw_spots()
    self._resize_img()

    self._calc_hist()
    self._resize_hist()

    self._display_img()
    self._display_hist()

    self._update_pixel_value()

    self.update()

  def _handle_box_outside_img(self, box: Box) -> None:
    """If a patch is outside the image, maybe the user entered a wrong value or
    the image size has been modified. Raising an exception as the DISVE can't
    run in this situation."""

    raise ValueError(f'The patch {box.get_patch()} does not fit in the '
                     f'image !')



# coding: utf-8

"""
Demonstration of how to create a custom out InOut in Crappy.

This InOut is intended to be used as a template, it doesn't actually act on any
real device.

No hardware required.
"""

import crappy
from time import sleep

# This class can be used as a starting point to create a new InOut object.
# To add it to crappy, make the imports relative (refer to any other inout),
# move the class to a file in crappy/inout and add the corresponding line
# in crappy/inout/__init__.py


class My_inout(crappy.inout.InOut):
  """A basic example of InOut object."""

  def __init__(self):
    # Do not forget to init InOut !
    super().__init__()

  def open(self):
    print("Opening device...")
    sleep(1)
    print("Device opened!")

  def close(self):
    print("Closing device...")
    sleep(.5)
    print("Device closed")

  def set_cmd(self, value):
    """This is the method that will be called to write the data."""

    print("Setting command to", value)


if __name__ == '__main__':
  flipflop = {
      'type': 'cyclic',
      'value1': 0,
      'condition1': 'delay=1',
      'value2': 1,
      'condition2': 'delay=1',
      'cycles': 100}

  gen = crappy.blocks.Generator([flipflop], cmd_label='cmd')

  io = crappy.blocks.IOBlock('My_inout', cmd_labels=['cmd'])

  crappy.link(gen, io)

  crappy.start()



# coding: utf-8

import numpy as np
from typing import List, Optional

from .block import Block


class Mean_block(Block):
  """This block computes the average values over a given delay of each label
  received, and returns them.

  It can take any number of inputs, provided that they share a common time
  label. If the same label (except time) is received from several blocks, it
  may lead to unexpected results.

  Warning:
    If the delay for averaging is too short compared with the looping frequency
    of the upstream blocks, this block may not always return the same number of
    labels ! This can cause errors in downstream blocks expecting a fixed
    number of labels.
  """

  def __init__(self,
               delay: float,
               time_label: str = 't(s)',
               out_labels: Optional[List[str]] = None,
               verbose: bool = False,
               freq: float = 50) -> None:
    """Sets the args and initializes the parent class.

    Args:
      delay: The averaged data will be sent each ``delay`` seconds.
      time_label: The label containing the time information.
      out_labels: If given, only the listed labels and the time will be
        returned. Otherwise, all of them are returned.
      verbose: If :obj:`True`, prints the looping frequency of the block.
      freq: The block will try to loop at this frequency.
    """

    super().__init__()
    self.verbose = verbose
    self.freq = freq

    self._delay = delay
    self._time_label = time_label
    self._out_labels = out_labels

  def prepare(self) -> None:
    """Initializes the buffer and the time counters."""

    self._buffer = {link: dict() for link in self.inputs}
    self._last_sent_t = -self._delay
    self._last_recv_t = 0

  def loop(self) -> None:
    """Receives all available data from the upstream blocks, and averages it
    and sends it if the time delay is reached."""

    # Receiving data from each incoming link
    for link in self.inputs:
      data = link.recv_chunk(blocking=False)

      if data is not None:
        # Updating the last received time attribute
        if self._time_label in data:
          self._last_recv_t = max(self._last_recv_t,
                                  data[self._time_label][-1])
          data.pop(self._time_label)

        # Storing the incoming data into dicts
        for label in data:
          if self._out_labels is None or label in self._out_labels:
            if label in self._buffer[link]:
              self._buffer[link][label].extend(data[label])
            else:
              self._buffer[link][label] = data[label]

    # Sending the mean value when the delay is reached
    if self._last_recv_t - self._last_sent_t > self._delay:
      ret = dict()

      # For each label of each dict, getting it mean value
      for dic in self._buffer.values():
        for label, values in dic.items():
          try:
            ret[label] = np.mean(values)
          except TypeError:
            ret[label] = values[-1]
        # Finally, clearing the buffer
        dic.clear()

      # Sending only if there was data to send
      if ret:
        ret[self._time_label] = np.mean((self._last_recv_t, self._last_sent_t))
        self.send(ret)
      self._last_sent_t = self._last_recv_t



# coding: utf-8

from typing import Optional, Tuple
import numpy as np

from .cameralink import Cl_camera
from .._global import OptionalModule
try:
  from . import clModule as Cl
except (ModuleNotFoundError, ImportError):
  Cl = OptionalModule("clModule")

format_to_num = {'8 bits': 0,
                 '10 bits': 1,
                 '12 bits': 2}
num_to_format = {val: key for key, val in format_to_num.items()}


class Jai8(Cl_camera):
  """This class supports Jai GO-5000-PMCL gray cameras.

  This one uses FullAreaGray8 module for maximum framerate.
  """

  def __init__(self) -> None:
    """"""

    super().__init__()

    self.add_scale_setting('width', 1, 2560, self._get_w, self._set_w, 2560)
    self.add_scale_setting('height', 1, 2048, self._get_h, self._set_h, 2048)
    self.add_scale_setting('exposure', 10, 800000, self._get_exp,
                           self._set_exp)

  def open(self,
           config_file: Optional[str] = None,
           camera_type: str = 'FullAreaGray8',
           **kwargs) -> None:
    """"""

    super().open(config_file=config_file,
                 camera_type=camera_type,
                 **kwargs)

    self.cap.serialWrite('TAGM=5\r\n')  # (default)
    self._set_format('8 bits')  # Set camera to 8 bits
    self.cap.set(Cl.FG_CAMERA_LINK_CAMTYP, 208)  # Set the input to 8 bits
    self.cap.set(Cl.FG_SENSORREADOUT,
                 0)  # Sets the correct framegrabber mode

  def _set_w(self, val: int) -> None:
    self.stopAcq()
    super()._set_w(val)
    self.cap.serialWrite('WTC={}\r\n'.format(val))
    self.startAcq()

  def _set_h(self, val: int) -> None:
    self.stopAcq()
    super()._set_h(val)
    self.cap.serialWrite('HTL={}\r\n'.format(val))
    self.startAcq()

  def _get_format(self) -> str:
    r = self.cap.serialWrite('BA?\r\n')
    return num_to_format[int(r[3])]

  def _set_format(self, val: str) -> None:
    self.cap.serialWrite('BA={}\r\n'.format(format_to_num[val]))

  def _set_exp(self, val: int) -> None:
    self.cap.serialWrite('PE={}\r\n'.format(val))

  def _get_exp(self) -> int:
    return int(self.cap.serialWrite('PE?\r\n').strip()[3:])


class Jai(Jai8):
  """This class allows the use of 10 and 12 bits mode for the Jai Cameras.
  Obviously, the framerate will be slower than the 8 bits version."""

  def __init__(self) -> None:
    """"""

    super().__init__()
    self.add_choice_setting('data_format', ('10 bits', '12 bits'),
                            self._get_format, self._set_format, '12 bits')

  def open(self,
           camera_type: str = 'MediumAreaGray16',
           **kwargs) -> None:
    """"""

    super().open(camera_type=camera_type,
                 **kwargs)
    # dual tap (default does not allow 12 bits)
    self.cap.serialWrite('TAGM=1\r\n')
    self._set_format('12 bits')  # 12 bits
    self.cap.set(Cl.FG_CAMERA_LINK_CAMTYP, 212)  # Set the input to 12 bits
    self.cap.set(Cl.FG_SENSORREADOUT, 7)  # Sets the correct framegrabber mode

  def get_image(self) -> Tuple[float, np.ndarray]:
    """"""

    t, frame = super().get_image()
    return t, frame >> 4



# coding: utf-8

from typing import Optional
from .inout import InOut
from .._global import OptionalModule

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")


class Gpio_pwm(InOut):
  """ Class for performing PWM on a Raspberry Pi.

  The Gpio_pwn InOut block is meant for controlling a PWM output from a
  Raspberry Pi GPIO. It allows to control the duty cycle, the frequency, or
  both at the same time. When controlling both, the duty cycle should be first
  and the frequency second in the list of inputs.

  Warning:
    Only works on Raspberry Pi !
  """

  def __init__(self,
               pin_out: int,
               duty_cycle: Optional[float] = None,
               frequency: Optional[float] = None) -> None:
    """Checks the arguments validity.

    Args:
      pin_out (:obj:`int`): The GPIO pin to be controlled (BCM convention).
      duty_cycle (:obj:`float`): If provided, sets a fixed duty cycle for the
        entire assay. Only the frequency can then be tuned. If not provided,
        the block will expect the duty cycle values to be given as the first
        input. It will also start the PWM with a duty cycle of `0%` before the
        first value is received and set.
      frequency (:obj:`float`): If provided, sets a fixed PWM frequency for the
        entire assay. Only the duty cycle can then be tuned. If not provided,
        the block will expect the frequency values to be given as the first
        input if the ``duty_cycle`` argument is provided, or else as the second
        input. It will also start the PWM with a frequency of `10kHz` before
        the first value is received and set.

    Note:
      - ``duty_cycle``:
        The duty cycle inputs are clamped between `0` and `100`.

      - ``frequency``:
        The frequency inputs are clamped between `10Hz` and `1MhZ`. Sending
        other values to the bloc doesn't raise any error, but the assay may not
        run as expected.

      - **Hardware PWM pins**:
        On the Raspberry Pi 4, only the GPIO pins `12`, `13`, `18` and `19`
        support hardware PWM. Trying to get a PWM output from other pins might
        work but may decrease the available frequency range.
    """

    super().__init__()

    if pin_out not in range(2, 28):
      raise ValueError("pin_out should be an integer between 2 and 28")
    else:
      self._pin_out = pin_out

    if frequency is not None:
      if not 10 <= frequency < 1000000:
        raise ValueError("frequency should be between 100Hz and 1MHz")
    self._frequency = frequency

    if duty_cycle is not None:
      if not 0 <= duty_cycle <= 100:
        raise ValueError("Duty cycle should be positive and not exceed 100%")
    self._duty_cycle = duty_cycle

  def open(self) -> None:
    """Sets the GPIOs and starts the PWM."""

    # Setting the GPIOs
    GPIO.setmode(GPIO.BCM)
    GPIO.setup(self._pin_out, GPIO.OUT)

    # Setting to user frequency if provided, or else to 10kHz
    if self._frequency is not None:
      self._pwm = GPIO.PWM(self._pin_out, self._frequency)
    else:
      self._pwm = GPIO.PWM(self._pin_out, 10000)

    # Setting to user duty cycle if provided, or else to 0%
    if self._duty_cycle is not None:
      self._pwm.start(self._duty_cycle)
    else:
      self._pwm.start(0)

  def set_cmd(self, *cmd: float) -> None:
    """Modifies the PWM frequency and/or duty cycle.

    Args:
      *cmd (:obj:`float`): Values of duty cycle and/or frequency to set
    """

    cmd = list(*cmd)
    # If both frequency and duty cycle are fixed by the user, nothing to do
    if self._duty_cycle is not None and self._frequency is not None:
      pass

    # If only frequency is fixed, setting the duty cycle
    elif self._frequency is not None:
      dc = min(100, max(0, cmd[0]))
      self._pwm.ChangeDutyCycle(dc)

    # If only the duty cycle is fixed, setting the frequency
    elif self._duty_cycle is not None:
      freq = min(1000000, max(10, cmd[0]))
      self._pwm.ChangeFrequency(freq)

    # If neither duty cycle nor frequency are fixed, setting both
    else:
      dc = min(100, max(0, cmd[0]))
      freq = min(1000000, max(10, cmd[1]))
      self._pwm.ChangeFrequency(freq)
      self._pwm.ChangeDutyCycle(dc)

  def close(self) -> None:
    """Stops PWM and releases GPIOs."""

    self._pwm.stop()
    GPIO.cleanup()



# coding: utf-8

from sys import platform
from multiprocessing import Process, Pipe
from time import sleep, time, localtime, strftime
from weakref import WeakSet
from pickle import UnpicklingError
from typing import Union, Optional, List, Dict, Any

from ..links import Link
from .._global import CrappyStop

import subprocess

# Todo:
#  Add a clean way to stop the blocks, using the keyboard or a button
#  Rethink the communication between blocks when stopping
#  Switch from dicts to structured arrays
#  Make sure that a variable number of labels never raises errors
#  Replace lists with tuples in arguments


def renice(pid: int, niceness: int) -> None:
  """Function to renice a process.

  Warning:
    Only works in Linux.

  Note:
    The user must be allowed to use ``sudo`` to renice with a negative value.
    May thus ask for a password for negative values.
  """

  if niceness < 0:
    subprocess.call(['sudo', 'renice', str(niceness), '-p', str(pid)])
  else:
    subprocess.call(['renice', str(niceness), '-p', str(pid)])


class Block(Process):
  """This represent a Crappy block, it must be parent of all the blocks."""

  instances = WeakSet()

  def __init__(self) -> None:
    Process.__init__(self)
    # Block.instances.append(self)
    self.outputs: List[Link] = []
    self.inputs: List[Link] = []
    # This pipe allows to send 2 essential signals:
    # pipe1->pipe2 is to start the main function and set t0
    # pipe2->pipe1 to set process status to the parent
    self.pipe1, self.pipe2 = Pipe()
    self._status = "idle"
    self.niceness = 0
    self.labels = []

    self._all_last_values = None
    self._last_values = None

  def __new__(cls, *args, **kwargs) -> Process:
    instance = super().__new__(cls)
    Block.instances.add(instance)
    return instance

  @classmethod
  def reset(cls) -> None:
    cls.instances = WeakSet()

  def run(self) -> None:
    self.status = "initializing"
    try:
      self.prepare()
      self.status = "ready"
      # Wait for parent to tell me to start the main
      self.t0 = self.pipe2.recv()
      if self.t0 < 0:
        try:
          self.finish()
        except (Exception,):
          pass
        return
      self.status = "running"
      self.begin()
      self._MB_last_t = time()
      self._MB_last_FPS = self._MB_last_t
      self._MB_loops = 0
      self.main()
      self.status = "done"
    except CrappyStop:
      print("[%r] Encountered CrappyStop Exception, terminating" % self)
      self.status = "done"
      self.stop_all()
    except KeyboardInterrupt:
      print("[%r] Keyboard interrupt received" % self)
    except Exception as e:
      print("[%r] Exception caught:" % self, e)
      try:
        self.finish()
      except (Exception,):
        pass
      self.status = "error"
      sleep(1)  # To let downstream blocks process the data and avoid loss
      self.stop_all()
      raise
    self.finish()
    self.status = "done"

  @classmethod
  def get_status(cls) -> List[str]:
    return [x.status for x in cls.instances]

  @classmethod
  def all_are(cls, s: str) -> bool:
    """Returns :obj:`True` only if all processes status are `s`."""

    lst = cls.get_status()
    return len(set(lst)) == 1 and s in lst

  @classmethod
  def renice_all(cls, high_prio: bool = True, **_) -> None:
    """Will renice all the blocks processes according to ``block.niceness``
    value.

    Note:
      If ``high_prio`` is :obj:`False`, blocks with a negative niceness value
      will be ignored.

      This is to avoid asking for the ``sudo`` password since only root can
      lower the niceness of processes.
    """

    if "win" in platform:
      # Not supported on Windows yet
      return
    for b in cls.instances:
      if b.niceness < 0 and high_prio or b.niceness > 0:
        print("[renice] Renicing", b.pid, "to", b.niceness)
        renice(b.pid, b.niceness)

  @classmethod
  def prepare_all(cls, verbose: bool = True) -> None:
    """Starts all the blocks processes (``block.prepare``), but not the main
    loop."""

    if verbose:
      def vprint(*args):
        print("[prepare]", *args)
    else:
      def vprint(*_):
        return
    vprint("Starting the blocks...")
    for instance in cls.instances:
      vprint("Starting", instance)
      instance.start()
      vprint("Started, PID:", instance.pid)
    vprint("All processes are started.")

  @classmethod
  def launch_all(cls,
                 t0: Optional[float] = None,
                 verbose: bool = True,
                 bg: bool = False) -> None:
    if verbose:
      def vprint(*args):
        print("[launch]", *args)
    else:
      def vprint(*_):
        return
    if not cls.all_are('ready'):
      vprint("Waiting for all blocks to be ready...")
    while not cls.all_are('ready'):
      sleep(.1)
      if not all([i in ['ready', 'initializing', 'idle']
                  for i in cls.get_status()]):
        print("Crappy failed to start!")
        for i in cls.instances:
          if i.status in ['ready', 'initializing']:
            i.launch(-1)
        cls.stop_all()
        return
        # raise RuntimeError("Crappy failed to start!")
    vprint("All blocks ready, let's go !")
    if not t0:
      t0 = time()
    vprint("Setting t0 to", strftime("%d %b %Y, %H:%M:%S", localtime(t0)))
    for instance in cls.instances:
      instance.launch(t0)
    t1 = time()
    vprint("All blocks loop started. It took", (t1 - t0) * 1000, "ms")
    if bg:
      return
    try:
      # Keep running
      lst = cls.get_status()
      while not ("done" in lst or "error" in lst):
        lst = cls.get_status()
        sleep(1)
    except KeyboardInterrupt:
      print("Main process got keyboard interrupt!")
      cls.stop_all()
      # It will automatically propagate to the blocks processes
    if not cls.all_are('running'):
      print('Waiting for all processes to finish')
    t = time()
    while 'running' in cls.get_status() and time() - t < 3:
      sleep(.1)
    if cls.all_are('done'):
      print("Crappy terminated gracefully")
    else:
      print("Crappy terminated, blocks status:")
      for b in cls.instances:
        print(b, b.status)

  @classmethod
  def start_all(cls,
                t0: float = None,
                verbose: bool = True,
                bg: bool = False,
                high_prio: bool = False) -> None:
    cls.prepare_all(verbose)
    if high_prio and any([b.niceness < 0 for b in cls.instances]):
      print("[start] High prio: root permission needed to renice")
    cls.renice_all(high_prio, verbose=verbose)
    cls.launch_all(t0, verbose, bg)

  @classmethod
  def stop_all(cls, verbose: bool = True) -> None:
    """Stops all the blocks (``crappy.stop``)."""

    if verbose:
      def vprint(*args):
        print("[stop]", *args)
    else:
      def vprint(*_):
        return
    vprint("Stopping the blocks...")
    for instance in cls.instances:
      if instance.status == 'running':
        vprint("Stopping", instance)
        instance.stop()
    vprint("All blocks are stopped.")

  def begin(self) -> None:
    """If :meth:`main` is not overridden, this method will be called first,
    before entering the main loop."""

    pass

  def finish(self) -> None:
    """If :meth:`main` is not overridden, this method will be called upon exit
    or after a crash."""

    pass

  def loop(self) -> None:
    raise NotImplementedError('You must override loop or main in' + str(self))

  def main(self) -> None:
    """This is where you define the main loop of the block.

    Important:
      If not overridden, will raise an error.
    """

    while not self.pipe2.poll():
      self.loop()
      self.handle_freq()
    print("[%r] Got stop signal, interrupting..." % self)

  def handle_freq(self) -> None:
    """For block with a given number of `loops/s` (use ``freq`` attribute to
    set it)."""

    self._MB_loops += 1
    t = time()
    if hasattr(self, 'freq') and self.freq:
      d = self._MB_last_t + 1 / self.freq - t
      while d > 0:
        t = time()
        d = self._MB_last_t + 1 / self.freq - t
        sleep(max(0, d / 2 - 2e-3))  # Ugly, yet simple and pretty efficient
    self._MB_last_t = t
    if hasattr(self, 'verbose') and self.verbose and \
            self._MB_last_t - self._MB_last_FPS > 2:
      print("[%r] loops/s:" % self,
            self._MB_loops / (self._MB_last_t - self._MB_last_FPS))
      self._MB_loops = 0
      self._MB_last_FPS = self._MB_last_t

  def launch(self, t0: float) -> None:
    """To start the :meth:`main` method, will call :meth:`Process.start` if
    needed.

    Once the process is started, calling launch will set the starting time and
    actually start the main method.

    Args:
      t0 (:obj:`float`): Time to set as starting time of the block (mandatory,
        in seconds after epoch).
    """

    if self.status == "idle":
      print(self, ": Called launch on unprepared process!")
      self.start()
    self.pipe1.send(t0)  # asking to start main in the process

  @property
  def status(self) -> str:
    """Returns the status of the block, from the process itself or the parent.

    It can be:

      - `"idle"`: Block not started yet.
      - `"initializing"`: :meth:`start` was called and :meth:`prepare` is not
        over yet.
      - `"ready"`: :meth:`prepare` is over, waiting to start :meth:`main` by
        calling :meth:`launch`.
      - `"running"`: :meth:`main` is running.
      - `"done"`: :meth:`main` is over.
      - `"error"`: An error occurred and the block stopped.
    """

    if self.pid is not None:
      while self.pipe1.poll():
        try:
          self._status = self.pipe1.recv()
        except (EOFError, UnpicklingError):
          if self._status == 'running':
            self._status = 'done'
      # If another process tries to get the status
      if 'win' not in platform:
        self.pipe2.send(self._status)

        # Somehow the previous line makes crappy hang on Windows, no idea why
        # It is not critical, but it means that process status can now only
        # be read from the process itself and ONE other process.
        # Luckily, only the parent (the main process) needs the status for now
    return self._status

  @status.setter
  def status(self, s: str) -> None:
    assert self.pid is not None, "Cannot set status from outside of the " \
                                 "process!"
    self.pipe2.send(s)
    self._status = s

  def prepare(self) -> None:
    """This will be run when creating the process, but before the actual start.

    The first code to be run in the new process, will only be called once and
    before the actual start of the main launch of the blocks.

    It can remain empty and do nothing.
    """

    pass

  def send(self, data: Union[Dict[str, float], List[float]]) -> None:
    """To send the data to all blocks downstream.

    Send has 2 ways to operate. You can either build the :obj:`dict` yourself,
    or you can define ``self.labels`` (usually time first) and call send with a
    :obj:`list`. It will then map them to the :obj:`dict`.

    Note:
      ONLY :obj:`dict` can go through links.
    """

    if isinstance(data, dict):
      pass
    elif isinstance(data, list):
      if not self.labels:
        raise IOError("trying to send data as a list but no labels are "
                      "specified ! Please add a self.labels attribute.")
      data = dict(zip(self.labels, data))
    elif data == 'stop':
      pass
    else:
      raise IOError("Trying to send a " + str(type(data)) + " in a link!")
    for o in self.outputs:
      o.send(data)

  def recv_all(self) -> Dict[str, list]:
    """Receives new data from all the inputs (not as chunks).

    It will simply call :meth:`Pipe.recv` on all non empty links and return a
    single :obj:`dict`.

    Important:
      If the same label comes from multiple links, it may be overridden !
    """

    r = {}
    for i in self.inputs:
      if i.poll():
        r.update(i.recv())
    return r

  def poll(self) -> bool:
    """Tells if any input link has pending data

    Returns True if l.poll() is True for any input link l.
    """

    return any((link.poll for link in self.inputs))

  def recv_all_last(self) -> Dict[str, float]:
    """Like recv_all, but drops older data to return only the latest value

    This method avoids Pipe congestion that can be induced by recv_all when the
    receiving block is slower than the block upstream
    """

    ret = {}
    for link in self.inputs:
      data = link.recv_last(blocking=False)
      if data is not None:
        ret.update(data)
    return ret

  def get_last(self, blocking: bool = True) -> Dict[str, Any]:
    """To get the latest value of each labels from all inputs.

    Warning:
      Unlike the ``recv`` methods of :ref:`Link`, this method is NOT guaranteed
      to return all the data going through the links!

      It is meant to get the latest values, discarding all the previous ones
      (for a displayer for example).

      Its mode of operation is completely different since it can operate on
      multiple inputs at once.

    Note:
      The first call may be blocking until it receives data, all the others
      will return instantaneously, giving the latest known reading.
    """

    # Initializing the buffer
    if self._last_values is None:
      self._last_values = [dict() for _ in self.inputs]

    for link, values in zip(self.inputs, self._last_values):
      # If blocking is True, acquiring a sample from each link at the beginning
      if not values and blocking:
        values.update(link.recv(blocking=True))

      # Then updating the buffers with the last received values
      data = link.recv_last(blocking=False)
      if data is not None:
        values.update(data)

    # Finally, gathering data from all the buffers and returning it
    ret = {}
    for values in self._last_values:
      ret.update(values)
    return ret

  def get_all_last(self, blocking: bool = True) -> Dict[str, list]:
    """To get the data from all links of the block.

    It is almost the same as :meth:`get_last`, but will return all the data
    that goes through the links (as :obj:`list`).

    Also, if multiple links have the same label, only the last link's value
    will be kept.

    Args:
      blocking: If :obj:`True`, blocks until there's at least one value
        received for each label.
    """

    if self._all_last_values is None:
      self._all_last_values = {link: dict() for link in self.inputs}

    for link in self.inputs:
      new_values = link.recv_chunk(blocking=blocking)
      if new_values is not None:
        self._all_last_values[link].update(new_values)

    ret = dict()
    for link in self.inputs:
      ret.update(self._all_last_values[link])

    return ret

  def recv_all_delay(self,
                     delay: Optional[float] = None,
                     poll_delay: float = .1) -> List[Dict[str, list]]:
    """Method to wait for data, but continuously reading all the links to make
    sure they do not saturate.

    Args:
      delay: The method only returns after this delay (in seconds). If
        :obj:`None` or `0`, it returns after polling the pipes just once.
      poll_delay: The delay (in seconds) between two pipe polls. It is safer to
        keep it lower than 0.1s. Not used when ``delay`` is :obj:`None` or `0`.

    Return:
      A :obj:`list` where each entry is what would have been returned by
      :meth:`Link.recv_chunk` on each link.
    """

    def poll(inputs: List[Link],
             rcv: List[Dict[str, list]]) -> None:
      """Polls all the incoming links and saves the received values.

      Args:
        inputs: The upcoming links.
        rcv: The :obj:`list` storing the received values.
      """

      for link, dict_ in zip(inputs, rcv):
        if not link.poll():
          continue
        new = link.recv_chunk(blocking=False)
        if new is not None:
          for key, value in new.items():
            if key in dict_:
              dict_[key].extend(value)
            else:
              dict_[key] = value

    received = [{} for _ in self.inputs]

    # Just poll the pipes, read the data and return
    if not delay:
      poll(self.inputs, received)

    # Poll the pipes and read the data every poll_delay until delay has expired
    else:
      last_t = time()
      while True:
        sleep(max(0., poll_delay - time() + last_t))
        poll(self.inputs, received)
        if time() - last_t > delay:
          break

    return received

  def drop(self, num: Optional[Union[list, str]] = None) -> None:
    """Will clear the inputs of the blocks.

    This method performs like :meth:`get_last`, but returns :obj:`None`
    instantly.
    """

    if num is None:
      num = range(len(self.inputs))
    elif not isinstance(num, list):
      num = [num]
    for n in num:
      self.inputs[n].clear()

  def add_output(self, out: Link) -> None:
    """Adds a :ref:`Link` as an output."""

    self.outputs.append(out)

  def add_input(self, in_: Link) -> None:
    """Adds a :ref:`Link` as an input."""

    self.inputs.append(in_)

  def stop(self) -> None:
    if self.status != 'running':
      return
    print('[%r] Stopping' % self)
    self.pipe1.send(0)
    for i in self.inputs:
      i.send('stop')
    for i in range(10):
      if self.status == "done":
        break
      sleep(.05)
    # if self.status != "done":
    if self.status not in ['done', 'idle', 'error']:
      print('[%r] Could not stop properly, terminating' % self)
      try:
        self.terminate()
      except (Exception,):
        pass
    else:
      print("[%r] Stopped correctly" % self)

  def __repr__(self) -> str:
    return str(type(self)) + " (" + str(self.pid or "Not running") + ")"



# coding: utf-8

from time import time, sleep
from typing import Union, Optional, List
from .inout import InOut
from ..tool import ft232h_server as ft232h, Usb_server
from .._global import OptionalModule

try:
  import smbus2
except (ModuleNotFoundError, ImportError):
  smbus2 = OptionalModule("smbus2")

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

# Register Map
NAU7802_Scale_Registers = {'PU_CTRL': 0x00,
                           'CTRL1': 0x01,
                           'CTRL2': 0x02,
                           'OCAL1_B2': 0x03,
                           'OCAL1_B1': 0x04,
                           'OCAL1_B0': 0x05,
                           'GCAL1_B3': 0x06,
                           'GCAL1_B2': 0x07,
                           'GCAL1_B1': 0x08,
                           'GCAL1_B0': 0x09,
                           'OCAL2_B2': 0x0A,
                           'OCAL2_B1': 0x0B,
                           'OCAL2_B0': 0x0C,
                           'GCAL2_B3': 0x0D,
                           'GCAL2_B2': 0x0E,
                           'GCAL2_B1': 0x0F,
                           'GCAL2_B0': 0x10,
                           'I2C_CONTROL': 0x11,
                           'ADCO_B2': 0x12,
                           'ADCO_B1': 0x13,
                           'ADCO_B0': 0x14,
                           'ADC': 0x15,
                           'OTP_B1': 0x16,
                           'OTP_B0': 0x17,
                           'PGA': 0x1B,
                           'PGA_PWR': 0x1C,
                           'DEVICE_REV': 0x1F}

# Bits within the PU_CTRL register
NAU7802_PU_CTRL_Bits = {'PU_CTRL_RR': 0,
                        'PU_CTRL_PUD': 1,
                        'PU_CTRL_PUA': 2,
                        'PU_CTRL_PUR': 3,
                        'PU_CTRL_CS': 4,
                        'PU_CTRL_CR': 5,
                        'PU_CTRL_OSCS': 6,
                        'PU_CTRL_AVDDS': 7}

# Bits within the CTRL2 register
NAU7802_CTRL2_Bits = {'CTRL2_CALMOD': 0,
                      'CTRL2_CALS': 2,
                      'CTRL2_CAL_ERROR': 3,
                      'CTRL2_CRS': 4,
                      'CTRL2_CHS': 7}

# Bits within the PGA PWR register
NAU7802_PGA_PWR_Bits = {'PGA_PWR_PGA_CURR': 0,
                        'PGA_PWR_ADC_CURR': 2,
                        'PGA_PWR_MSTR_BIAS_CURR': 4,
                        'PGA_PWR_PGA_CAP_EN': 7}

# Allowed Low drop out regulator voltages
NAU7802_LDO_Values = {2.4: 0b111,
                      2.7: 0b110,
                      3.0: 0b101,
                      3.3: 0b100,
                      3.6: 0b011,
                      3.9: 0b010,
                      4.2: 0b001,
                      4.5: 0b000}

# Allowed gains
NAU7802_Gain_Values = {1: 0b000,
                       2: 0b001,
                       4: 0b010,
                       8: 0b011,
                       16: 0b100,
                       32: 0b101,
                       64: 0b110,
                       128: 0b111}

# Allowed samples per second
NAU7802_SPS_Values = {10: 0b000,
                      20: 0b001,
                      40: 0b010,
                      80: 0b011,
                      320: 0b111}

# Calibration state
NAU7802_Cal_Status = {'CAL_SUCCESS': 0,
                      'CAL_IN_PROGRESS': 1,
                      'CAL_FAILURE': 2}

NAU7802_Backends = ['Pi4', 'ft232h']

NAU7802_VREF = 3.3


class Nau7802(Usb_server, InOut):
  """Class for controlling Sparkfun's NAU7802 load cell conditioner.

  The Nau7802 InOut block is meant for reading output values from a NAU7802
  load cell conditioner, using the I2C protocol. The output is in Volts by
  default, but can be converted to Newtons using ``gain`` and ``offset``.
  """

  def __init__(self,
               backend: str,
               i2c_port: int = 1,
               device_address: int = 0x2A,
               gain_hardware: int = 128,
               sample_rate: int = 80,
               int_pin: Optional[Union[str, int]] = None,
               gain: float = 1,
               offset: float = 0,
               ft232h_ser_num: Optional[str] = None) -> None:
    """Checks the validity of the arguments..

    Args:
      backend (:obj:`str`): The backend for communicating with the NAU7802.
        Should be one of:
        ::

          'Pi4', 'ft232h'

        to use the device from a Raspberry Pi or from a PC through an FT232H
        USB to I2C converter, respectively. See :ref:`Crappy for embedded
        hardware` for details.
      i2c_port (:obj:`int`, optional): The I2C port over which the NAU7802
        should communicate. On most Raspberry Pi models the default I2C port is
        `1`.
      device_address (:obj:`int`, optional): The I2C address of the NAU7802. It
        is impossible to change this address, so it is not possible to have
        several NAU7802 on the same i2c bus.
      gain_hardware (:obj:`int`, optional): The gain to be used by the
        programmable gain amplifier. Setting a high gain allows reading small
        voltages with a better precision, but it might saturate the sensor for
        higher voltages. Available gains are:
        ::

          1, 2, 4, 8, 16, 32, 64, 128

      sample_rate (:obj:`int`, optional): The sample rate for data conversion.
        The higher the rate, the greater the noise. Available sample rates are:
        ::

          10, 20, 40, 80, 320

      int_pin (:obj:`int` or :obj:`str`, optional): Optionally, reads the end
        of conversion signal from a GPIO rather than from an I2C message.
        Speeds up the reading and decreases the traffic on the bus, but
        requires one extra wire. With the backend `'Pi4'`, give the index of
        the GPIO in BCM convention. With the `'ft232h'` backend, give the name
        of the GPIO in the format `Dx` or `Cx`.
      gain (:obj:`float`, optional): Allows to tune the output value according
        to the formula:
        ::

          output = gain * tension + offset.

      offset (:obj:`float`, optional): Allows to tune the output value
        according to the formula:
        ::

          output = gain * tension + offset.

      ft232h_ser_num (:obj:`str`, optional): If backend is `'ft232h'`, the
        serial number of the ft232h to use for communication.
    """

    if backend not in NAU7802_Backends:
      raise ValueError("backend should be in {}".format(NAU7802_Backends))

    Usb_server.__init__(self,
                        serial_nr=ft232h_ser_num if ft232h_ser_num else '',
                        backend=backend)
    InOut.__init__(self)
    current_file, block_number, command_file, answer_file, block_lock, \
        current_lock = super().start_server()

    if backend == 'Pi4':
      self._bus = smbus2.SMBus(i2c_port)
    else:
      self._bus = ft232h(mode='I2C',
                         block_number=block_number,
                         current_file=current_file,
                         command_file=command_file,
                         answer_file=answer_file,
                         block_lock=block_lock,
                         current_lock=current_lock,
                         serial_nr=ft232h_ser_num)
    self._device_address = device_address
    self._backend = backend

    if gain_hardware not in NAU7802_Gain_Values:
      raise ValueError("gain_hardware should be in {}".format(list(
        NAU7802_Gain_Values.keys())))
    else:
      self._gain_hardware = gain_hardware

    if sample_rate not in NAU7802_SPS_Values:
      raise ValueError("sample_rate should be in {}".format(list(
        NAU7802_SPS_Values.keys())))
    else:
      self._sample_rate = sample_rate

    if int_pin is not None:
      if backend == 'ft232h' and not isinstance(int_pin, str):
        raise TypeError('int_pin should be a string when using the ft232h '
                        'backend !')
      elif backend == 'Pi4' and not isinstance(int_pin, int):
        raise TypeError('int_pin should be an int when using the Pi4 '
                        'backend !')
    self._int_pin = int_pin

    self._gain = gain
    self._offset = offset

  def open(self) -> None:
    """Sets the I2C communication and device."""

    if not self._is_connected():
      raise IOError("The NAU7802 is not connected")

    # Resetting the device
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_RR'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)
    sleep(0.001)
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_RR'],
                  NAU7802_Scale_Registers['PU_CTRL'], 0)

    # Powering up the device - takes approx 200µs
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUD'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUA'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)

    t0 = time()
    while not self._get_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUR'],
                            NAU7802_Scale_Registers['PU_CTRL']):
      if time() - t0 > 0.1:
        raise TimeoutError("Waited too long for the device to power up !")

    # Setting the Low Drop Out voltage to 3.3V and setting the gain
    value = NAU7802_Gain_Values[self._gain_hardware]
    self._bus.write_byte_data(self._device_address,
                              NAU7802_Scale_Registers['CTRL1'], value)
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_AVDDS'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)

    # Setting the sample rate
    self._bus.write_byte_data(self._device_address,
                              NAU7802_Scale_Registers['CTRL2'],
                              NAU7802_SPS_Values[self._sample_rate] << 4)

    # Turning off CLK_CHP
    self._bus.write_byte_data(self._device_address,
                              NAU7802_Scale_Registers['ADC'], 0x30)

    # Enabling 330pF decoupling cap on channel 2
    self._set_bit(NAU7802_PGA_PWR_Bits['PGA_PWR_PGA_CAP_EN'],
                  NAU7802_Scale_Registers['PGA_PWR'], 1)

    # Re-calibrating the analog front-end
    self._set_bit(NAU7802_CTRL2_Bits['CTRL2_CALS'],
                  NAU7802_Scale_Registers['CTRL2'], 1)

    t0 = time()
    while self._cal_afe_status() != NAU7802_Cal_Status['CAL_SUCCESS']:
      if time() - t0 > 1:
        raise TimeoutError("Waited too long for the calibration to occur !")
      if self._cal_afe_status() == NAU7802_Cal_Status['CAL_FAILURE']:
        raise IOError("Calibration failed !")

    if self._backend == 'Pi4' and self._int_pin is not None:
      GPIO.setmode(GPIO.BCM)
      GPIO.setup(self._int_pin, GPIO.IN)

  def get_data(self) -> List[float]:
    """Reads the registers containing the conversion result.

    The output is in Volts by default, and can be converted to Newtons using
    gain and offset.

    Returns:
      :obj:`list`: A list containing the timeframe and the output value
    """

    # Waiting for data to be ready
    t0 = time()
    while not self._data_available():
      if time() - t0 > 0.5:
        raise TimeoutError('Waited too long for data to be ready !')

    out = [time()]

    # Reading the output data
    block = self._bus.read_i2c_block_data(self._device_address,
                                          NAU7802_Scale_Registers['ADCO_B2'],
                                          3)
    value_raw = (block[0] << 16) | (block[1] << 8) | block[2]

    # Converting raw data into Volts or Newtons
    if block[0] >> 7:
      value_raw -= 2 ** 24
    value = value_raw / 2 ** 23 * 0.5 * NAU7802_VREF / self._gain_hardware
    out.append(self._offset + self._gain * value)
    return out

  def close(self) -> None:
    """Powers down the device."""

    # Powering down the device
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUD'],
                  NAU7802_Scale_Registers['PU_CTRL'], 0)
    sleep(0.001)
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUA'],
                  NAU7802_Scale_Registers['PU_CTRL'], 0)
    self._bus.close()

    if self._backend == 'Pi4' and self._int_pin is not None:
      GPIO.cleanup()

  def _is_connected(self) -> bool:
    """Tries reading a byte from the device.

    Returns:
      :obj:`bool`: :obj:`True` if reading was successful, else :obj:`False`
    """

    try:
      self._bus.read_byte(self._device_address)
      return True
    except IOError:
      return False

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._int_pin is None:
      return self._get_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_CR'],
                           NAU7802_Scale_Registers['PU_CTRL'])
    # EOC signal from a GPIO
    elif self._backend == 'ft232h':
      return bool(self._bus.get_gpio(self._int_pin))
    else:
      return bool(GPIO.input(self._int_pin))

  def _set_bit(self,
               bit_number: int,
               register_address: int,
               bit: int) -> None:
    """Sets a given bit in the specified register.

    Args:
      bit_number (:obj:`int`): Position of the bit in the register
      register_address (:obj:`int`): Index of the register
      bit (:obj:`int`): Value of the bit
    """

    value = self._bus.read_i2c_block_data(self._device_address,
                                          register_address, 1)[0]
    if bit:
      value |= (1 << bit_number)
    else:
      value &= ~(1 << bit_number)
    self._bus.write_byte_data(self._device_address, register_address, value)

  def _get_bit(self,
               bit_number: int,
               register_address: int) -> bool:
    """Reads a given bit in the specified register.

    Args:
      bit_number (:obj:`int`): Position of the bit in the register
      register_address (:obj:`int`): Index of the register

    Returns:
      :obj:`bool`: True if the bit value is 1, else False
    """

    value = self._bus.read_i2c_block_data(self._device_address,
                                          register_address, 1)[0]
    value = value >> bit_number & 1
    return bool(value)

  def _cal_afe_status(self) -> int:
    """Reads the calibration status bits.

    Returns:
      :obj:`int`: The int value corresponding to the current calibration status
    """

    if self._get_bit(NAU7802_CTRL2_Bits['CTRL2_CAL_ERROR'],
                     NAU7802_Scale_Registers['CTRL2']):
      return NAU7802_Cal_Status['CAL_FAILURE']
    elif self._get_bit(NAU7802_CTRL2_Bits['CTRL2_CALS'],
                       NAU7802_Scale_Registers['CTRL2']):
      return NAU7802_Cal_Status['CAL_IN_PROGRESS']
    else:
      return NAU7802_Cal_Status['CAL_SUCCESS']



# coding:utf-8

class OptionalModule:
  """Placeholder for optional dependencies when not installed

  Will print a message and raise an error when trying to use them
  """

  def __init__(self, module_name, message=None):
    self.mname = module_name
    if message is not None:
      self.message = message
    else:
      self.message = """The module {} is necessary to use this functionality
Please install it and try again""".format(self.mname)

  def __getattr__(self, _):
    print("Missing module: {}".format(self.mname))
    print(self.message)
    raise RuntimeError(self.message)

  def __call__(self, *_, **__):
    print("Missing module: {}".format(self.mname))
    print(self.message)
    raise RuntimeError(self.message)


class CrappyStop(Exception):
  """Error to raise when Crappy is terminating"""

  pass


class DefinitionError(Exception):
    """Error to raise when classes are not defined correctly"""

    def __init__(self, msg=""):
        self.msg = msg

    def __str__(self):
        return self.msg



# coding: utf-8

from typing import Callable, Optional, Tuple, Union, Any, Dict
from time import sleep
import numpy as np

from .._global import DefinitionError

nbr_type = Union[int, float]


class MetaCam(type):
  """Metaclass ensuring that two cameras don't have the same name, and that all
  cameras define the required methods. Also keeps track of all the Camera
  classes, including the custom user-defined ones."""

  classes = {}

  def __new__(mcs, name: str, bases: tuple, dct: dict) -> type:
    return super().__new__(mcs, name, bases, dct)

  def __init__(cls, name: str, bases: tuple, dct: dict) -> None:
    super().__init__(name, bases, dct)

    # Checking that a Camera with the same name doesn't already exist
    if name in cls.classes:
      raise DefinitionError(f"The {name} class is already defined !")

    # Otherwise, saving the class
    if name != "Camera":
      cls.classes[name] = cls


class Cam_setting:
  """Base class for each camera setting.

  It is meant to be subclassed and should not be used as is.
  """

  def __init__(self,
               name: str,
               getter: Callable[[], Any],
               setter: Callable[[Any], None],
               default: Any) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
    """

    # Attributes shared by all the settings
    self.name = name
    self.default = default
    self.type = type(default)

    # Attributes used in the GUI
    self.tk_var = None
    self.tk_obj = None

    # Attributes for internal use only
    self._value_no_getter = default
    self._getter = getter
    self._setter = setter

  @property
  def value(self) -> Any:
    """Returns the current value of the setting, by calling the getter if one
    was provided or else by returning the stored value."""

    if self._getter is not None:
      return self._getter()
    else:
      return self._value_no_getter

  @value.setter
  def value(self, val: Any) -> None:
    self._value_no_getter = val
    if self._setter is not None:
      self._setter(val)

    if self.value != val:
      print(f'[Cam settings] Could not set {self.name} to {val}, the value '
            f'is {self.value} !')


class Cam_bool_setting(Cam_setting):
  """Camera setting that can only be :obj:`True` or :obj:`False`."""

  def __init__(self,
               name: str,
               getter: Optional[Callable[[], bool]] = None,
               setter: Optional[Callable[[bool], None]] = None,
               default: bool = True) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
    """

    super().__init__(name, getter, setter, default)


class Cam_scale_setting(Cam_setting):
  """Camera setting that can take any value between a lower and an upper
  boundary.

  This class can handle settings that should only take :obj:`int` values as
  well as settings that can take :obj:`float` value.
  """

  def __init__(self,
               name: str,
               lowest: nbr_type,
               highest: nbr_type,
               getter: Optional[Callable[[], nbr_type]] = None,
               setter: Optional[Callable[[nbr_type], None]] = None,
               default: Optional[nbr_type] = None) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      lowest: The lower boundary for the setting values.
      highest: The upper boundary for the setting values.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
    """

    self.lowest = lowest
    self.highest = highest
    self.type = int if isinstance(lowest + highest, int) else float

    if default is None:
      default = self.type((lowest + highest) / 2)

    super().__init__(name, getter, setter, default)

  @property
  def value(self) -> nbr_type:
    """Returns the current value of the setting, by calling the getter if one
    was provided or else by returning the stored value."""

    if self._getter is not None:
      return self.type(min(max(self._getter(), self.lowest), self.highest))
    else:
      return self.type(self._value_no_getter)

  @value.setter
  def value(self, val: nbr_type) -> None:
    val = min(max(val, self.lowest), self.highest)

    self._value_no_getter = self.type(val)
    if self._setter is not None:
      self._setter(self.type(val))

    if self.value != val:
      print(f'[Cam settings] Could not set {self.name} to {val}, the value '
            f'is {self.value} !')


class Cam_choice_setting(Cam_setting):
  """Camera setting that can take any value from a predefined list of
  values."""

  def __init__(self,
               name: str,
               choices: Tuple[str, ...],
               getter: Optional[Callable[[], str]] = None,
               setter: Optional[Callable[[str], None]] = None,
               default: Optional[str] = None) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      choices: A tuple listing the possible values for the setting.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
    """

    self.choices = choices

    if default is None:
      default = choices[0]

    super().__init__(name, getter, setter, default)


class Camera(metaclass=MetaCam):
  """Base class for every camera object.

  It contains all the methods shared by these classes and sets MetaCam as their
  metaclass.
  """

  def __init__(self) -> None:
    """Simply sets the dict containing the settings."""

    self.settings: Dict[str, Cam_setting] = dict()

  def open(self, **kwargs) -> None:
    """This method should initialize the connection to the camera, configure
    the camera, and start the image acquisition.

    This method also takes as arguments all the kwargs that were passed to the
    Camera block but not used by it. Some may be used directly, e.g. for
    choosing which camera to open out of several possible ones, and the others
    should indicate values to set for available settings. It is fine not to
    provide any values for the settings here, as each setting has a default.

    To effectively set the setting values, the method :meth:`set_all` has to
    be called at the end of open (e.g. ``self.set_all(**kwargs)``). This is
    true even if no value to set was given in the kwargs. If it is not called,
    the settings won't actually be set on the camera.

    If some camera settings require values from the camera for their
    instantiation (e.g.
    ``self.add_setting(..., highest=self.cam.max_width(), ...)``), they should
    be instantiated here. And of course before calling :meth:`set_all`.
    """

    self.set_all(**kwargs)

  def get_image(self) -> Optional[Union[Tuple[Dict[str, Any], float],
                                        np.ndarray]]:
    """Acquires an image and returns it along with its metadata or timestamp.

    It is also fine for this method to return :obj:`None`. The image should be
    returned as a numpy array, and the metadata as a :obj:`dict` or the
    timestamp as a :obj:`float`. The keys of the metadata dictionary should
    preferably be valid Exif tags, so that the metadata can be embedded into
    the image file when saving.

    In order for the recording of images to run, the metadata dict must
    contain at least the ``'t(s)'`` and ``''ImageUniqueID''`` keys, whose
    values should be the timestamp when the frame was acquired (as returned by
    ``time.time()``) and the frame number as an :obj:`int`.
    """

    print(f"WARNING ! The get_img method is not defined for the Camera "
          f"{type(self).__name__} !\nNo image can be acquired if this method "
          f"isn't defined !")
    sleep(1)
    return

  def close(self) -> None:
    """This method should perform any action required for properly stopping the
    image acquisition and/or closing the connection to the camera.

    This step is usually extremely important in order for the camera resources
    to be released. Otherwise, it might be impossible to re-open the camera
    from Crappy without resetting the hardware connection with it.
    """

    ...

  def add_bool_setting(self,
                       name: str,
                       getter: Optional[Callable[[], bool]] = None,
                       setter: Optional[Callable[[bool], None]] = None,
                       default: bool = True) -> None:
    """Adds a boolean setting, whose value is either :obj:`True` or
    :obj:`False`.

    Args:
      name: The name of the setting, that will be displayed in the GUI and can
        be used to directly get the value of the setting by calling
        ``self.<name>``
      getter: The method for getting the current value of the setting. If not
        given, the returned value is simply the last one that was set.
      setter: The method for setting the current value of the setting. If not
        given, the value to be set is simply stored.
      default: The default value to assign to the setting.
    """

    if name in self.settings:
      raise ValueError('This setting already exists !')
    self.settings[name] = Cam_bool_setting(name, getter, setter, default)

  def add_scale_setting(self,
                        name: str,
                        lowest: nbr_type,
                        highest: nbr_type,
                        getter: Optional[Callable[[], nbr_type]] = None,
                        setter: Optional[Callable[[nbr_type], None]] = None,
                        default: Optional[nbr_type] = None) -> None:
    """Adds a scale setting, whose value is an :obj:`int` or a :obj:`float`
    clamped between two boundaries.

    Note:
      If any of ``lowest`` or ``highest`` is a :obj:`float`, then the setting
      is considered to be of type float and can take float values. Otherwise,
      it is considered of type int and can only take integer values.

    Args:
      name: The name of the setting, that will be displayed in the GUI and can
        be used to directly get the value of the setting by calling
        ``self.<name>``
      lowest: The lowest possible value for the setting.
      highest: The highest possible value for the setting.
      getter: The method for getting the current value of the setting. If not
        given, the returned value is simply the last one that was set.
      setter: The method for setting the current value of the setting. If not
        given, the value to be set is simply stored.
      default: The default value to assign to the setting. If not given, will
        be the average of ``lowest`` and ``highest``.
    """

    if name in self.settings:
      raise ValueError('This setting already exists !')
    self.settings[name] = Cam_scale_setting(name, lowest, highest, getter,
                                            setter, default)

  def add_choice_setting(self,
                         name: str,
                         choices: Tuple[str, ...],
                         getter: Optional[Callable[[], str]] = None,
                         setter: Optional[Callable[[str], None]] = None,
                         default: Optional[str] = None) -> None:
    """Adds a choice setting, that can take a limited number of predefined
    :obj:`str` values.

    Args:
      name: The name of the setting, that will be displayed in the GUI and can
        be used to directly get the value of the setting by calling
        ``self.<name>``
      choices: A :obj:`tuple` containing the possible values for the setting.
      getter: The method for getting the current value of the setting. If not
        given, the returned value is simply the last one that was set.
      setter: The method for setting the current value of the setting. If not
        given, the value to be set is simply stored.
      default: The default value to assign to the setting. If not given, will
        be the fist item in ``choices``.
    """

    if name in self.settings:
      raise ValueError('This setting already exists !')
    self.settings[name] = Cam_choice_setting(name, choices, getter, setter,
                                             default)

  def set_all(self, **kwargs) -> None:
    """Checks if the kwargs are valid, sets them, and for settings that are not
    in kwargs sets them to their default value."""

    unexpected = tuple(kwarg for kwarg in kwargs if kwarg not in self.settings)
    if unexpected:
      raise ValueError(f'Unexpected argument(s) {", ".join(unexpected)} for '
                       f'camera {type(self).__name__}.')

    for name, setting in self.settings.items():
      setting.value = kwargs[name] if name in kwargs else setting.default

  def __getattr__(self, item: str) -> Any:
    """Method for getting the value of a setting directly by calling
    ``self.<setting name>``.

    It is called in case __getattribute__ doesn't work properly, and tries to
    return the corresponding setting value."""

    try:
      return self.settings[item].value
    except (AttributeError, KeyError):
      raise AttributeError(f'No attribute nor setting named {item}')

  def __setattr__(self, key: str, val: Any) -> None:
    """Method for setting the value of a setting directly by calling
    ``self.<setting name> = <value>``."""

    if key != 'settings' and key in self.settings:
      self.settings[key].value = val
    else:
      super().__setattr__(key, val)



# coding: utf-8

"""
Detailed example showing a videoextensometry-driven tensile test.

This example uses a Labjack T7 board to send the position command to a tensile
machine. This was tested on an Instron 5882, but any tensile machine taking an
external command can be used. A camera is used to measure the strain of the
sample and save the images. Different levels of strain are applied and the
sample is relaxed between each step.

Required hardware:
  - Tensile machine able to take an external command
  - DAQ board (here a Labjack T7)
  - A Camera
"""

import crappy

if __name__ == "__main__":
  SPEED = 5  # mm/min
  FORCE_GAIN = 100  # N/V
  POS_GAIN = 5  # mm/V

  # This path will be a bit more complicated:
  # Load until the strain is reached, unload until F(N)<1
  paths = []
  for strain in [.25, .5, .75, 1., 1.5, 2]:
    paths.append({'type': 'constant',
                  'value': SPEED / 60,
                  'condition': 'Exx(%)>{}'.format(strain)})
    # Go up to this level
    paths.append({'type': 'constant',
                  'value': -SPEED / 60,
                  'condition': 'F(N)<1'})  # Go down to F=1N

  # Just like the previous example, we create the generator with our new path
  gen = crappy.blocks.Generator(paths, cmd_label='cmd')

  # The DAQ board, this section is identical
  force = {'name': 'AIN0', 'gain': FORCE_GAIN}
  pos = {'name': 'AIN1', 'gain': POS_GAIN, 'make_zero': True}
  cmd = {'name': 'TDAC0', 'gain': POS_GAIN}
  daq = crappy.blocks.IOBlock('Labjack_t7', channels=[force, pos, cmd],
                              labels=['t(s)', 'F(N)', 'Position(mm)'],
                              cmd_labels=['cmd'])

  crappy.link(gen, daq)
  # And we ALSO need to link the daq to the generator because it takes
  # decisions based on the values of the force
  # crappy.link(a,b) and crappy.link(b,a) are NOT equivalent because links are
  # unidirectional ! Data only flows from the first block to the second
  crappy.link(daq, gen)
  # It is perfectly fine to make "loops" by linking in both directions
  # like this

  # Recorder for the data from the machine
  rec_daq = crappy.blocks.Recorder('results_daq.csv')
  crappy.link(daq, rec_daq)

  # Now let's add our now sensor: videoextensometry
  # We can specify arguments specific to the camera in the block
  # save_folder asks the block to save the images
  # It can be useful for further processing after the test
  ve = crappy.blocks.Video_extenso('Webcam', save_folder='img/')
  # When the program will start, this will open a windows to preview the
  # video from the camera, adjust the settings and select the markers.

  # Now we must link this block to the generator
  crappy.link(ve, gen)
  # and to a recorder
  rec_ve = crappy.blocks.Recorder('results_ve.csv')
  crappy.link(ve, rec_ve)
  # Quick remark on recorders: we used two separate blocks for daq and ve.
  # This is because they can only take a SINGLE INPUT. Because the input blocks
  # may run at different frequencies, it is not possible to build a csv file
  # simply from two or more separate sources.
  # We could interpolate them in the same
  # timebase (see block Multiplex) but this means loosing the raw data. It is
  # much better to perform this interpolation in post-processing if necessary

  # Our force against time plot
  graph_f = crappy.blocks.Grapher(('t(s)', 'F(N)'))
  # Adding a Mean modifier to average the points
  # It lowers the frequency and makes the graph smoother
  crappy.link(daq, graph_f, modifier=crappy.modifier.Mean(10))

  # Let's add a second graph to plot strain against time
  graph_s = crappy.blocks.Grapher(('t(s)', 'Exx(%)'))
  crappy.link(ve, graph_s)

  # We can now start the program
  crappy.start()



# coding: utf-8

from .modifier import Modifier, MetaModifier

from .demux import Demux
from .differentiate import Diff
from .integrate import Integrate
from .mean import Mean
from .median import Median
from .moving_avg import Moving_avg
from .moving_med import Moving_med
from .offset import Offset
from .trig_on_change import Trig_on_change
from .trig_on_value import Trig_on_value

modifier_list = MetaModifier.classes



# coding: utf-8

from typing import Optional, Dict, Any
from .modifier import Modifier


class Trig_on_change(Modifier):
  """Modifier passing the data to the downstream block only when the value of
  a given label changes.

  It also transmits the first received data. Can be used to trig a block upon
  change of a label value.
  """

  def __init__(self, label: str) -> None:
    """Sets the args and initializes the parent class.

    Args:
      label: The name of the label to monitor.
    """

    super().__init__()
    self._label = label
    self._last = None

  def evaluate(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Compares the received value with the last sent one, and if they're
    different sends the received data and stores the latest value."""

    # Storing the first received value and returning the data
    if self._last is None:
      self._last = data[self._label]
      return data

    # Returning the data if the label value is different from the stored value
    if data[self._label] != self._last:
      self._last = data[self._label]
      return data



# coding: utf-8
"""
Base file for tests using videoextensometry and a marker following actuator.
"""

from time import ctime
import crappy

out_gain = 1 / 30  # V/mm
gains = [50, 1 / out_gain]  # N/V mm/V

timestamp = ctime()[:-5].replace(" ", "_").replace(":", "_")
save_path = "./" + timestamp + "/"

# VideoExtenso and Autodrive blocks
ve = crappy.blocks.Video_extenso(camera='Ximea_cv', show_image=True,
                                 white_spots=False, max_fps=30)

ad = crappy.blocks.AutoDrive(
    actuator={'name': 'CM_drive', 'port': '/dev/ttyUSB0'}, direction='X-')

graph_extenso = crappy.blocks.Grapher(('t(s)', 'Exx(%)'), ('t(s)', 'Eyy(%)'))

rec_extenso = crappy.blocks.Recorder(save_path+"extenso.csv",
                                     labels=['t(s)', 'Exx(%)', 'Eyy(%)'])

# Linking them
crappy.link(ve, graph_extenso)
crappy.link(ve, rec_extenso)
crappy.link(ve, ad)

# Labjack
lj = crappy.blocks.IOBlock("Labjack_t7", channels=[
  {'name': 'AIN0', 'gain': gains[0], 'make_zero':True},
  {'name': 'AIN1', 'gain': gains[1], 'make_zero':True},
  {'name': 'TDAC0', 'gain': out_gain}],
                           labels=['t(s)', 'F(N)', 'x(mm)'],
                           cmd_labels=['cmd'])

# Graph
graph_sensors = crappy.blocks.Grapher(('t(s)', 'F(N)'), ('t(s)', 'x(mm)'))
crappy.link(lj, graph_sensors, modifier=crappy.modifier.Mean(10))

# Generator
g = crappy.blocks.Generator(path=[
  {'type': 'cyclic_ramp', 'condition1': 'Exx(%)>20',
   'speed1': 20 / 60, 'condition2': 'F(N)<.1', 'speed2': -20 / 60,
   'cycles': 5}, ])
rec_sensors = crappy.blocks.Recorder(save_path + "sensors.csv",
                                     labels=['t(s)', 'F(N)', 'x(mm)'])

# Linking the generator to all the blocks
crappy.link(ve, g)
crappy.link(lj, g)
crappy.link(g, lj)
crappy.link(lj, rec_sensors)

crappy.start()



# coding: utf-8

from typing import Optional, Union
from .inout import InOut
from .._global import OptionalModule
from ..tool import ft232h_server as ft232h, Usb_server, ft232h_pin_nr

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

try:
  import board
except (ModuleNotFoundError, ImportError):
  board = OptionalModule('board', 'Blinka is necessary to access the GPIO')

try:
  import digitalio
except (ImportError, ModuleNotFoundError):
  digitalio = OptionalModule('digitalio',
                             'Blinka is necessary to access the GPIOs')

gpio_switch_backends = ['Pi4', 'ft232h', 'blinka']


class Gpio_switch(Usb_server, InOut):
  """Class for setting a GPIO high or low.

  The Gpio_switch InOut block is meant for switching a GPIO high or low
  according to the input signal value. When the input signal is `1` the
  GPIO is turned high, when the signal is `0` it is turned low. Any value other
  than `0` and `1` raises an error.
  """

  def __init__(self,
               pin_out: Union[int, str],
               backend: str,
               ft232h_ser_num: Optional[int] = None) -> None:
    """Checks the argument validity.

    Args:
      pin_out: The GPIO pin to be controlled. On Raspberry Pi, should be an
        integer corresponding to a GPIO in BCM convention. On FT232H, should be
        a string corresponding to the name of a GPIO. With the `'blinka'`
        backend, should be a string holding the name of the pin. Refer to
        blinka's specific documentation for each board for more information.
      backend: Should be one of :
        ::

          'Pi4', 'blinka', 'ft232h'

        The `'Pi4'` backend only works on the Raspberry Pis. The `'blinka'`
        backend requires installing Adafruit's modules, but is compatible with
        and maintained on a wide variety of boards. The `'ft232h'` backend
        allows controlling the GPIO from a PC using Adafruit's FT232H USB to
        I2C adapter. See :ref:`Crappy for embedded hardware` for details.
      ft232h_ser_num: If backend is `'ft232h'`, the serial number of the FT232H
        to use for communication.
    """

    # Checking that the backend is valid
    if not isinstance(backend, str) or backend not in gpio_switch_backends:
      raise ValueError("backend should be in {}".format(gpio_switch_backends))
    self._backend = backend

    # Checking that the given pin is valid for the selected backend
    if backend == 'blinka' and not hasattr(board, pin_out):
      raise TypeError(f'{pin_out} is not a valid pin using the blinka backend '
                      f'on this board !')
    elif backend == 'ft232h' and pin_out not in ft232h_pin_nr:
      raise TypeError(f'{pin_out} is not a valid pin for the ft232h backend !')
    elif backend == 'Pi4' and (not isinstance(pin_out, int) or
                               pin_out not in range(2, 28)):
      raise ValueError('pin_out should be an integer between 2 and 28 when '
                       'using the Pi4 backend !')

    # Starting the USB server (doesn't do anything if backend is not ft232h)
    Usb_server.__init__(self,
                        serial_nr=ft232h_ser_num if ft232h_ser_num else '',
                        backend=backend)
    InOut.__init__(self)
    current_file, block_number, command_file, answer_file, block_lock, \
        current_lock = super().start_server()

    # Instantiating the pin object
    if backend == 'Pi4':
      self._pin_out = pin_out
    elif backend == 'ft232h':
      self._pin_out = pin_out
      self._ft232h = ft232h(mode='GPIO_only',
                            block_number=block_number,
                            current_file=current_file,
                            command_file=command_file,
                            answer_file=answer_file,
                            block_lock=block_lock,
                            current_lock=current_lock,
                            serial_nr=ft232h_ser_num)
    elif backend == 'blinka':
      self._pin_out = digitalio.DigitalInOut(getattr(board, pin_out))

  def open(self) -> None:
    """Sets the GPIO."""

    if self._backend == 'Pi4':
      GPIO.setmode(GPIO.BCM)
      GPIO.setup(self._pin_out, GPIO.OUT)

    elif self._backend == 'blinka':
      self._pin_out.direction = digitalio.Direction.OUTPUT

  def set_cmd(self, *cmd: int) -> None:
    """Drives the GPIO according to the command.

    Args:
      cmd (:obj:`int`): 1 for driving the GPIO high, 0 for driving it low
    """

    if cmd[0] not in [0, 1]:
      raise ValueError("The GPIO input can only be 0 or 1")

    if self._backend == 'Pi4':
      GPIO.output(self._pin_out, cmd[0])
    elif self._backend == 'blinka':
      self._pin_out.value = cmd[0]
    elif self._backend == 'ft232h':
      self._ft232h.set_gpio(self._pin_out, cmd[0])

  def close(self) -> None:
    """Releases the GPIO."""

    if self._backend == 'Pi4':
      GPIO.cleanup()
    elif self._backend == 'blinka':
      self._pin_out.deinit()
    elif self._backend == 'ft232h':
      self._ft232h.close()



# coding: utf-8

"""
Example demonstrating the use of National instrument daq boards.

Required hardware:
  - NI DAQ board with at least one analog input and one digital input
    compatible with the nidaqmx module.
"""

import crappy
from functools import partial


def intify(data):
  """Used to replace False by 0 and True by 1 in the digital inputs.

  Not doing so results in an error in the grapher.
  """

  for i, d in enumerate(data):
    if isinstance(d, bool):
      data[i] = int(d)
  return data


def change_name(data, prev: str, new: str):
  """Modifier for changing the name of a label."""

  val = data[prev]
  del data[prev]
  data[new] = val

  return data


if __name__ == "__main__":
  gen = crappy.blocks.Generator([dict(type='cyclic', value1=0, value2=1,
                                      condition1="delay=1",
                                      condition2="delay=1")], repeat=True)
  io = crappy.blocks.IOBlock("Nidaqmx", device="Dev2",
                             channels=[dict(name='ai0'), dict(name='di0'),
                                       dict(name='ao0'), dict(name='do1')],
                             samplerate=100,
                             labels=['t(s)', 'ai0', 'di0'],
                             cmd_labels=['cmd', 'cmd2'])
  crappy.link(gen, io)
  crappy.link(gen, io, modifier=partial(change_name, prev='cmd', new='cmd2'))
  graph = crappy.blocks.Grapher(('t(s)', 'di0'), ('t(s)', 'ai0'))
  crappy.link(io, graph, modifier=intify)
  crappy.start()



# coding: utf-8

from time import time

from .block import Block
from .._global import OptionalModule

try:
  import tkinter as tk
except (ModuleNotFoundError, ImportError):
  tk = OptionalModule("tkinter")


class GUI(Block):
  """This block allows the user to send a signal upon clicking on a button in
  a graphical user interface.

  It sends an integer value, that starts from `0` and is incremented every time
  the user clicks on the button.
  """

  def __init__(self,
               send_0: bool = False,
               label: str = 'step',
               time_label: str = 't(s)',
               freq: float = 50,
               spam: bool = False,) -> None:
    """Sets the args and initializes the parent class."""

    Block.__init__(self)
    self.freq = freq
    self.spam = spam
    self.labels = [time_label, label]

    self._send_0 = send_0

  def prepare(self) -> None:
    """Creates the graphical interface and sets its layout and callbacks."""

    self._root = tk.Tk()
    self._root.title("GUI block")
    self._root.resizable(False, False)

    self._step = tk.IntVar()
    self._step.trace_add('write', self._update_text)
    self._text = tk.StringVar(value=f'step: {self._step.get()}')

    self._label = tk.Label(self._root, textvariable=self._text)
    self._label.pack(padx=7, pady=7)

    self._button = tk.Button(self._root,
                             text='Next step',
                             command=self._next_step)
    self._button.pack(padx=25, pady=7)

    self._root.update()

  def begin(self) -> None:
    """Sends the value of the first step (`0`) if required."""

    if self._send_0:
      self.send([time() - self.t0, self._step.get()])

  def loop(self) -> None:
    """Updates the interface, and sends the current step value if spam is
    :obj:`True`. """

    try:
      self._root.update()
    except tk.TclError:
      return

    if self.spam:
      self.send([time() - self.t0, self._step.get()])

  def finish(self) -> None:
    """Closes the interface window."""

    try:
      self._root.destroy()
    except tk.TclError:
      pass

  def _update_text(self, _, __, ___) -> None:
    """Simply updates the displayed text."""

    self._text.set(f'step: {self._step.get()}')

  def _next_step(self) -> None:
    """Increments the step counter and sends the corresponding signal."""

    self._step.set(self._step.get() + 1)
    self.send([time() - self.t0, self._step.get()])
